import sys, os
from math import log, exp, log10
from datetime import date
from time import clock, localtime
import random
from copy import deepcopy
import utilities


class SingleDescriptor:
	def __init__(self, name, weight=0., norm=2., distanceCalc='euclidean', limit=False, simultaneous=None, origin='DESCRIPTORS', neededBy=['target', 'corpus'], packagename=None):
		self.singleNumberDescriptors = ['effDur-seg', 'peakTime-seg', 'MIDIPitch-seg', 'percentageInFile-seg']
		self.neverRecalculate = ['zeroCross', 'f0', 'peakamp', 'peakfrq']

		self.name = name
		self.weight = weight
		self.norm = norm
		self.distanceCalc = distanceCalc
		self.limit = limit
		self.neededBy = neededBy
		self.origin = origin
		self.energyweight = False
		self.packagename = packagename

		if name.find('-slope-seg') != -1:
			self.type = 'slope-regression'
			self.seg = True
		elif name.find('-mean-seg') != -1:
			self.type = 'mean'
			self.seg = True
		elif name.find('-seg') != -1:
			self.type = 'segmented'
			self.seg = True
		elif name.find('-odf-') != -1:
			self.type = 'onsetdetection'
			self.seg = False
		elif name.find('-delta') != -1:
			self.type = 'delta'
			self.seg = False
		elif name.find('-deltadelta') != -1:
			self.type = 'deltadelta'
			self.seg = False
		else:
			self.type = 'rawsdif'
			self.seg = False

		if self.seg:
			if self.type == 'slope-regression': self.seg_method = 'slope'
			elif self.type == 'mean': self.seg_method = 'mean'
			elif self.name in ["power-seg", "rms-seg"]: self.seg_method = 'max'
			elif self.name in ["f0-seg"]: self.seg_method = 'median'
			elif self.name in self.singleNumberDescriptors: self.seg_method = 'single_number'
			else: self.seg_method = 'weighted_mean'


		# get the "parent descriptors" which this descriptor depend upon
		namesplit = self.name.split('-')
		self.parents = []
		if self.name in self.singleNumberDescriptors: pass
		elif self.type in ['slope-regression', 'onsetdetection', 'delta', 'deltadelta', 'mean']:
			self.parents = [namesplit[0]]
		elif self.type == 'segmented':
			self.parents = []
			cnt = 1
			while True:
				if cnt > len(namesplit)-1: break
				self.parents.append( '-'.join(namesplit[:cnt]) )
				cnt += 1
	###########
	def __repr__(self):
		return self.name

		
class DescriptorManager:
	def __init__(self):
		# special names that indicate combinations of desciptors
		self.packagedDesciptors = ['mfccEnv', 'mfccFrq', 'chromas', 'peakfrqs', 'loudestpeak', 'crests', 'flatnesses', 'harmonictristimuluses', 'perceptualtristimuluses', 'cluster_cents']
		# don't ever weight by energy
		self.describesAmplitude = ['power']  
		# these descriptors can't be mixed, seg methods included
		self.neverMix = ['zeroCross', 'f0', 'effDur-seg', 'dur-seg', 'MIDIPitch-seg', 'percentageInFile-seg', 'peakamp', 'peakfrq'] # matches partial strings!
		#
		self.singleNumberDescriptors = ['effDur-seg', 'peakTime-seg', 'MIDIPitch-seg', 'percentageInFile-seg', 'temporalIncrease-seg', 'temporalDecrease-seg', 'logAttackTime-seg', 'temporalCentroid-seg']
		# here are things not attempted to evalute by SdifInterfaceescripor
		self.descriptorsLoadIgnore = ['loudestpeak']
		# here get added metrics needed for analysis and subtraction but not called for by the user
		self.internalMetrics = ['power', 'peakTime-seg', 'power-seg', 'power-mean-seg', 'effDur-seg']
		# here get added metrics needed for the corpus
		self.internalSrcMetrics = ['MIDIPitch-seg']  # used for musical writing/score outputs
		# check to see if midipitch method requires a descriptor
		for c in ops.CORPUS:
			if type(c) == dict and c.has_key('MIDIPitch') and c['MIDIPitch'] in ['centroid-seg', 'f0-seg'] and c['MIDIPitch'] not in self.internalMetrics:
				self.internalMetrics.append(c['MIDIPitch'])
		if ops.CORPUS_GLOBAL_ATTRIBUTE_DEFAULTS.has_key('MIDIPitch') and ops.CORPUS_GLOBAL_ATTRIBUTE_DEFAULTS['MIDIPitch'] in ['centroid-seg', 'f0-seg'] and ops.CORPUS_GLOBAL_ATTRIBUTE_DEFAULTS['MIDIPitch'] not in self.internalMetrics:
				self.internalMetrics.append(ops.CORPUS_GLOBAL_ATTRIBUTE_DEFAULTS['MIDIPitch'])
		if ops.CORPUS_GLOBAL_ATTRIBUTES.has_key('MIDIPitch') and ops.CORPUS_GLOBAL_ATTRIBUTES['MIDIPitch'] in ['centroid-seg', 'f0-seg'] and ops.CORPUS_GLOBAL_ATTRIBUTES['MIDIPitch'] not in self.internalMetrics:
				self.internalMetrics.append(ops.CORPUS_GLOBAL_ATTRIBUTES['MIDIPitch'])
		# smart Transpose
		if ops.SMART_TRANSPOSE: self.internalMetrics.append('f0-seg')
		
		self.limitsFromCorpus = []
		# add any global limits to limitList
		if ops.CORPUS_GLOBAL_ATTRIBUTES.has_key('limit'):
			for key in ops.CORPUS_GLOBAL_ATTRIBUTES['limit']:
				if key not in self.limitsFromCorpus: self.limitsFromCorpus.append(key)
		# loop through corpus items and add limits to the limit list
		for entry in ops.CORPUS:
			if type(entry) != dict or not entry.has_key('limit'): continue
			for key in entry['limit']:
				if key not in self.limitsFromCorpus: self.limitsFromCorpus.append(key)
		##########################
		## NEW FILL DESCRIPTORS ##
		##########################
		self.master_list = []
		self.master_name_list = []
		
		##############################################
		## make desciptor similarity matching steps ##
		##############################################
		self.DESCRIPTORS = []
		if type(ops.DESCRIPTORS) in [tuple, list]:
			self.DESCRIPTORS = ops.DESCRIPTORS
		else: # doesn't conform to new method
			if len(ops.FIRST_PASS_DESCRIPTORS) > 0 and ops.FIRST_PASS_PERCENT > 0:
				self.DESCRIPTORS.append( (ops.FIRST_PASS_DESCRIPTORS, ops.SIMILARITY_FIRST_PASS_SEARCH_RESULT_TYPE, ops.SIMILARITY_FIRST_PASS_PERCENTAGE) )
			self.DESCRIPTORS.append( (ops.DESCRIPTORS, ops.SIMILARITY_DESCRIPTOR_SEARCH_RESULT_TYPE, ops.SIMILARITY_DESCRIPTOR_PERCENTAGE) )
		p.log("Using descriptor hierarchical search phases:", self.DESCRIPTORS)
		##############################################
		## make desciptor similarity matching steps ##
		##############################################		
		for sidx, (dicty, method, opt1) in enumerate(self.DESCRIPTORS):
			for desc in dicty:
				w, n, m = self.getUserParams(dicty[desc])
				self.addDescriptor(desc, weight=w, norm=n, distanceCalc=m, origin='SEARCH_PASS%i'%sidx, neededBy=['target', 'corpus'])

		for origin, dicty in [('TARGET_ONSET_DESCRIPTORS', ops.TARGET_ONSET_DESCRIPTORS)]:
			for desc in dicty:
				self.addDescriptor(desc, weight=dicty[desc], norm=None, distanceCalc=None, simultaneous='subtract', origin=origin, neededBy=['target'])

		for origin, dicty in [('LOAD_ADDITIONAL_DESCRIPTORS', ops.LOAD_ADDITIONAL_DESCRIPTORS), ('AUDIOGUIDE_INTERNAL', self.internalMetrics)]:
			for desc in dicty:
				self.addDescriptor(desc, weight=0, norm=None, distanceCalc=None, origin=origin, neededBy=['target', 'corpus'])

		for origin, dicty in [('AUDIOGUIDE_INTERNAL_CORPUS', self.internalSrcMetrics), ('ORDER_CORPUS_BY_DESCRIPTOR', ops.ORDER_CORPUS_BY_DESCRIPTOR), ('CORPUS_RESTRICT_DISCRIPTOR_IN_TIME', ops.CORPUS_RESTRICT_DISCRIPTOR_IN_TIME)]:
			for desc in dicty:
				self.addDescriptor(desc, weight=0, norm=None, distanceCalc=None, origin=origin, neededBy=['corpus'])

		for origin, dicty in [('LIMIT_CORPUS_BY_DESCRIPTOR', ops.LIMIT_CORPUS_BY_DESCRIPTOR), ('limitsFromCorpusOps', self.limitsFromCorpus)]:
			for desc in dicty:
				self.addDescriptor(desc, weight=0, norm=None, distanceCalc=None, limit=True, origin=origin, neededBy=['corpus'])

		# done with lists
		self.getlists()
		self.getsimcalclist()
	###########
	def addDescriptor(self, name, weight=0., norm=None, distanceCalc=None, limit=False, simultaneous=None, origin=None, neededBy=None):
		packagename, descriptorList = metrics.name2packageExpansion(name, weight, neededBy)
		for dname, weight, neededBy in descriptorList:
			self.master_list.append( SingleDescriptor(dname, weight=weight, norm=norm, limit=limit, distanceCalc=distanceCalc, origin=origin, neededBy=neededBy, packagename=packagename) )
			self.master_name_list.append(dname)
	###########
	def getUserParams(self, listy):
		''' interprets the dict entries, e.g. {mfcc: (1, 2, 'manhattan')}'''
		if type(listy) not in [tuple, list]: listy = [listy]
		weight = listy[0]
		norm = ops.NORMALIZE_DEFAULT_METHOD
		distanceMethod = ops.SIMILARITY_ARRAY_DISTANCE_DEFAULT
		if len(listy) == 2:
			norm = listy[1]
		if len(listy) == 3:
			distanceMethod = listy[2]
		return weight, norm, distanceMethod
	###########
	def addParents(self):		
		for d in self.master_list:
			for pname in d.parents:
				if pname in self.master_name_list: continue
				self.addDescriptor(pname, weight=0., norm=None, distanceCalc=None, origin='PARENT', neededBy=d.neededBy)
	###########
	def getDescriptorObjByName(self, name):
		return self.master_list[self.master_name_list.index(name)]
	###########
	def getlists(self):
		self.target_load = []
		self.target_seg = []
		self.target_onfly = []
		self.target_descriptor_subtract = []
		self.target_descriptor_mixture = []
		self.corpus_load = []
		self.corpus_limit = []
		self.corpus_onfly = []
		self.normalize_compute = []
		
		self.addParents()
		# do mixture list!
		for d in self.master_list:
			tmp_power_list = [] # because power descriptors must come last in the procession
			if d.origin.find('SEARCH_PASS') == -1: continue
			mix_me = True
			if d.weight == 0: mix_me = False
			for skip in self.neverMix:
				if d.name.find(skip) != -1: mix_me = False
			if not mix_me: continue
			descripts = []
			for pname in d.parents: descripts.append(self.getDescriptorObjByName(pname))
			descripts.append(d) # goes after parent descriptors
			# make sure power if last!
			if d.name in self.describesAmplitude: tmp_power_list.extend(descripts)
			else: self.target_descriptor_mixture.extend(descripts)
		if len(tmp_power_list) == 0:
			# we MUST have power in there, even if it is not in the distance metrics
			self.target_descriptor_mixture.append(self.getDescriptorObjByName('power'))
		else:
			self.target_descriptor_mixture.extend(tmp_power_list)
		
		# do subtraction list!
		for d in self.master_list:
			if d.origin not in ['TARGET_ONSET_DESCRIPTORS']: continue
			for string in self.describesAmplitude:
				if d.name.find(string) == -1: continue
				for pname in d.parents:					
					self.target_descriptor_subtract.append(self.getDescriptorObjByName(pname))
				self.target_descriptor_subtract.append(d)
		
		for d in self.master_list:
			if 'target' in d.neededBy:
				if d.type == 'rawsdif': self.target_load.append(d)
				elif d.type == 'segmented': self.target_seg.append(d)
				else: self.target_onfly.append(d)
			if 'corpus' in d.neededBy:
				if d.type == 'rawsdif': self.corpus_load.append(d)
				else: self.corpus_onfly.append(d)
#			if d.distanceCalc:
#				if d.origin == 'FIRST_PASS_DESCRIPTORS':
#					self.similarity_compute_first_pass.append(d)
#				elif d.origin == 'DESCRIPTORS':
#					self.similarity_compute_main.append(d)
			if d.norm != None:
				self.normalize_compute.append(d)
			if d.limit:
				self.corpus_limit.append(d)
		# feature modification
		self.target_feature_mod = []
		for (anal, modStr, modVal) in ops.TARGET_DESCRIPTORS_MOD:
			if anal not in self.master_name_list:
				print "ERROR, cannot use an descriptor in TARGET_DESCRIPTORS_MOD that is not used for distance calcuatios in DESCRIPTORS or FIRST_PASS_DESCRIPTORS."
				sys.exit(2)
			d = self.getDescriptorObjByName(anal)
			for pname in d.parents:
				pd = self.getDescriptorObjByName(pname)
				self.target_feature_mod.append((pd, modStr, modVal))
			self.target_feature_mod.append((d, modStr, modVal))


		# done
		self.target_load = set(self.target_load)
		self.target_seg = set(self.target_seg)
		self.target_onfly = set(self.target_onfly)
		self.corpus_load = set(self.corpus_load)
		self.corpus_onfly = set(self.corpus_onfly)
		self.normalize_compute = set(self.normalize_compute)
	###########
	def getsimcalclist(self):
		#print DESCRIPTORS
		''' makes a list of first pass, main'''
		self.simliarity_search_steps = []
		for sidx, step in enumerate(self.DESCRIPTORS):
			phase = {'array_descriptors': [], 'singlenumber_descriptors': []}
			for d in self.master_list:
				if d.origin == 'SEARCH_PASS%i'%sidx and d.weight > 0:
					if d.seg: phase['singlenumber_descriptors'].append(d)
					else: phase['array_descriptors'].append(d)
					phase['search_result_type'] = (step[1], step[2])
					if step[1] in ['closest', 'farthest']:
						phase['complete_results'] = False 
					else: phase['complete_results'] = True 
			self.simliarity_search_steps.append(phase)
	############################################################################
	def allDescriptorsList(self, SdifInterface):
		self.allSegmentMethods = []
		for name in SdifInterface.agDescriptToSdif.keys():
			add_me = True
			for string in self.singleNumberDescriptors:
				if string.find(name) != -1: add_me = False
			if add_me: self.allSegmentMethods.append(name+'-seg')
		# write a file if it doesn't exist!
		textPath = os.path.join(os.path.dirname(__file__), 'all-descriptors.txt')
		if os.path.exists(textPath): return
		fh = open(textPath, 'w')
		all_d = []
		single_n_d = self.singleNumberDescriptors[:]
		max_len = 0
		for d in SdifInterface.agDescriptToSdif.keys():
			if d.find('-seg') != -1:
				single_n_d.append(d)
				continue
			if len(d) > max_len: max_len = len(d)
			all_d.append(d)
		all_d.sort()
		col_size = max_len+6
		fh.write('--------------------------------------------------------------\n')
		fh.write('----------------------  SINGLE NUMBER  -----------------------\n')
		fh.write('These descriptors consist of a single number for a segment of sound.  they are useful in temporal modeling -- i.e., effDur-seg describes the effectuve duration of a trget or corpus segment.  note that "effDur" does not exist -- this is why it is called a single number descriptor\n\n')
		for d in single_n_d:
			fh.write('%s\n'%d)
		fh.write('\n\n')
		fh.write('--------------------------------------------------------------\n')
		fh.write('-------------------  PACKAGED DESCRIPTOR  --------------------\n')
		fh.write('These descriptors refer to a group of other descritors with (sometimes) differing weights.  They are provided for convenience.  Here, the package descriptor name is shown to equal a list of "child" descriptors plus their weights.\n\n')
		for d in self.packagedDesciptors:
			name, listy = metrics.name2packageExpansion(d, 1, ['target'])
			pack = []
			for l in listy:
				pack.append('(%s %2.2f)'%(l[0], l[1]))
			fh.write('%s = %s\n\n'%(d, ', '.join(pack)))
		fh.write('\n\n')
		fh.write('--------------------------------------------------------------\n')
		fh.write('-----------------------  TIME VARYING  ------------------------\n')
		fh.write('These descriptors vary in time since each analysis frame produces a different value.  Note that any descriptor here, for example "centroid", may have different ending appended to it alter its values.\n\n')
		for d in all_d:
			fh.write('%s -- also: %s-delta, %s-deltadelta, %s-seg, %s-slope-seg\n'%(d, d, d, d, d))
		fh.close()









################################################################################
################################################################################
################################################################################
################################################################################
################################################################################
class analData: # holds data and stuff for database members
	def __init__(self):
		self.pb = None # init pb printing
		#self.setSampleRate()
		self.frameSec = ops.DESCRIPTOR_HOP_SIZE_SEC

		self.scoresToRender = []
		self.allHandles = [] # a list of all sample handles
		self.sourceHandles = [] # a list of all database sample handles
		self.hannWindow = metrics.hannWin(2096)
		self.minSegLength = int(ops.TARGET_MIN_SEGMENT_LENGTH/self.frameSec)
		self.maxSegLength = int(ops.TARGET_MAX_SEGMENT_LENGTH/self.frameSec)
		self.negEnergy = 0
		self.analNormalizeDict = {}
		self.srcMaxPowers = []
		self.corpusLengthInSeconds = 0
		self.normPrintData = {}
		self.selectedSegmentTimes = []
		self.selectedSegments = [] # maintain a list of selected corpus segments
		self.activeSegments = [] # maintain a list of currently active corpus segments
		self.countRanOutOfSamples = 0
		self.logicForSegmentTermination = {'max overlapping segments': 0, 'max segments per onset': 0, 'below onset threshold': 0, 'out of valid samples': 0,}
		self.timerDict = {}
		self.limitAnalStats = {}
	############################################################################
	def timer(self, name, call):
		from time import clock
		if call == "set":
			self.timerDict[name] = clock()
			return
		elif call == "get":
			self.timerDict[name] = clock()-self.timerDict[name]
			return self.timerDict[name]
	############################################################################
	def printTimeData(self):
		postLevel = 2
		totalTime = 0
		for key, val in self.timerDict.iteritems():
			totalTime += val
		p.middle('Elapsed Time - Total '+utilities.trunc(totalTime, 2)+" seconds", postLevel)
		for key, val in self.timerDict.iteritems():
			p.post(str(key)+" -> ${RED}"+utilities.trunc(val, 2)+"${NORMAL} seconds", postLevel)
		p.post("", postLevel)
	############################################################################
	def segTimesToWindowedLength(self, h):
		if h.startSec == None: h.startSec = 0.
		if h.endSec == None: h.endSec = h.sfdur
		h.lengthInFrames = int((h.endSec-h.startSec)/ops.DESCRIPTOR_HOP_SIZE_SEC)
		if h.lengthInFrames < 1: # at least one frame!
			h.endSec = h.startSec+ops.DESCRIPTOR_HOP_SIZE_SEC
			h.lengthInFrames = 1
	############################################################################
#	def loadAnalsIRCAM(self, handle, anals):
#		#analsFromDB, analsOnFly	= self.findAnalsOnFly(anals)
#		#print "a"
#		segAnals, onFlyAnals = SdifInterface.SdifSetDescriptors(handle.filename, anals)
#		#print handle.filename, handle.startSec, handle.endSec
#		SdifInterface.getDescriptorSegment(handle.filename, handle.startSec, handle.endSec, handle.data['raw'])
#		#print "\n\tIRCAM MADE", handle.data['raw'].keys()
#		#print "\tREMAINING",invalidAnals
#		if handle.envelopeMask != None:
#			SdifInterface.maskAmpDescriptsWithSyntheticEnvelope(handle.data['raw'], handle.envelopeMask)
#		#print "f"
#		return segAnals, onFlyAnals
	############################################################################
	def segmentTimes(self, handle, sfdur):
		if handle.startSec in [None,"start","Start"]: handle.startSec = 0 # if no startSec set to 0
		if handle.endSec in [None,"end","End"]: handle.endSec = sfdur
		if handle.endSec > sfdur: handle.endSec = sfdur
	############################################################################
	def makeTargetAttenuationGrid(self, handle):
		#############################
		## TARGET GRID ATTENUATION ##
		#############################
		# test to see if we need to make a temporal grid overlayed on the target's amplitude
		if type(ops.TARGET_MAKE_GRID_ATTENUATION) == dict and handle.typed == 'target':
			# make articulation points with randomness if specified
			attackPointsFrames = []
			peakAmps = []
			t = 0
			while t+ops.TARGET_MAKE_GRID_ATTENUATION['duration'] < tgt.lengthInFrames*ag.frameSec:
				rndAdd = random.uniform(ops.TARGET_MAKE_GRID_ATTENUATION['randomnessSec']/-2., ops.TARGET_MAKE_GRID_ATTENUATION['randomnessSec']/2.)
				time = round((t+rndAdd)/ag.frameSec, 0)
				attackPointsFrames.append(time)
				peakAmps.append(tgt.data['raw']['power'][time])
				t += ops.TARGET_MAKE_GRID_ATTENUATION['period']
			peakMaxAmp = np.max(peakAmps)
			# make the maksing array for tgt power multiplication
			maskArray = np.zeros(tgt.lengthInFrames)
			for onset in attackPointsFrames:
				onset = int(onset)
				offset = int(onset+(ops.TARGET_MAKE_GRID_ATTENUATION['duration']/ag.frameSec))
				singleGrid = np.linspace(1., 0., num=offset-onset)
				# add exponential slopping
				singleGrid = np.power(singleGrid, ops.TARGET_MAKE_GRID_ATTENUATION['slope'])
#				print onset, offset
				maskArray[onset:offset] = singleGrid
				# if normalize, make all grid points have then same peak amp
				if ops.TARGET_MAKE_GRID_ATTENUATION['norm']:
					boost = peakMaxAmp/tgt.data['raw']['power'][onset]
					maskArray[onset:offset] *= boost
			# mask it
			tgt.data['raw']['power'] *= maskArray
	############################################################################
	def calculatePostSelectionError(self, s, e, selectedSegments):
		return 0.
		#for d in descriptorData.rvw
		error = 0.
		return error
	############################################################################
	def segAnal(self, handle, d, startFrame, endFrame):
		# a dictionary with (start,end) tuple as key and descriptor as value
		handle.data['raw'][d.name][startFrame] = metrics.analSegment(d, handle, startFrame, endFrame, "Seg", ag)[0]
	############################################################################
	def clearSegDict(self, pointerToDict, time, dur):
		for (st,ed) in pointerToDict.keys():
			if st >= time and ed <= time+dur:
				del pointerToDict[(st,ed)]
	############################################################################
	def updateActiveSegments(self, currentTime, length):
		activeSegments = []
		for segment in self.selectedSegments:
			if segment[0] < currentTime+length and segment[1] > currentTime: # include it!
				activeSegments.append(segment)
		return activeSegments	
	############################################################################
	def postNormalizePrint(self):
		printLevel = 2
		p.middle('Normalization Data', 2)
		for d in descriptorData.normalize_compute:#, data in self.normPrintData.iteritems():
			data = self.normPrintData[d.name]
			printString = ""
			printString += "${RED}"+d.name.upper()+"${NORMAL}: "
			if data['sign'] == '-': printString += "${GREEN}INVERTED${NORMAL} and "
			printString += "normalized "
			if data['percent'] != 100 and data['norm'] == 'between': printString += str(data['percent'])+"% "
			if data['norm'] == 'together':
				printString += "${YELLOW}COMBINDING${NORMAL} the TARGET and SRC data "
			elif data['norm'] == 'separate':
				printString += "${YELLOW}SEPARATING${NORMAL} the TARGET and SRC data "
			elif data['norm'] == 'between':
				printString += "${YELLOW}INTERPOLATING BETWEEN${NORMAL} the TARGET and SRC data "

			if data['percent'] != 100 and data['norm'] != 'between':
				printString += "${GREEN}"
				if data['percent'] > 100:
					printString += "; expanding at "
				elif data['percent'] < 100:
					printString += "; compressing at "
				printString += str(data['percent'])+"% "
				printString += "${NORMAL}"

			p.post(printString, printLevel)
			if self.targetNormCoeffs[d.name] != self.sourceNormCoeffs[d.name]:
				p.post("\tTARGET "+d.name+": mean->"+utilities.trunc(self.targetNormCoeffs[d.name][0], 3)+" std->"+utilities.trunc(self.targetNormCoeffs[d.name][1], 3)+"\tDATABASE "+d.name+": mean->"+utilities.trunc(self.sourceNormCoeffs[d.name][0], 3)+" std->"+utilities.trunc(self.sourceNormCoeffs[d.name][1], 3)+"", printLevel)
			else:
				p.post("\tTARGET+DATABASE "+d.name+": mean->"+utilities.trunc(self.sourceNormCoeffs[d.name][0], 1)+" std->"+utilities.trunc(self.sourceNormCoeffs[d.name][1], 3)+"", printLevel)
		p.post("", printLevel)




################################################################################
################################################################################
################################################################################
################################################################################
################################################################################
class targetData:
	def __init__(self, filename, startSec, endSec):
		exists, self.filename, fileType = path.test(filename, ops.TARGET_PATH) # user path, relative path
		if not exists:
			print "\n\nTarget file", filename, "not found!\n"
			utilities.exit(1)
		self.name = filename.split('/')[-1] # short name for printing
		self.typed = 'target'
		self.startSec = startSec
		self.endSec = endSec
		self.sfdur, self.sfchannels = SdifInterface.checkSdifResource( self, self.filename )
		self.scaleDb = ops.TARGET_SCALE_DB
		ag.segmentTimes(self, self.sfdur)
		ag.segTimesToWindowedLength(self)
		self.onsetTime = 0
		self.offsetTime = 0
		self.envelopeMask = None
		self.timeInFrames = 0 # timekeeper
		self.segmentCounter = 0 # count target segments
		self.lastTgtSegTimes = (None, None)
		self.lastNoteTimeInFrames = 0 # for segment tracking
		self.needMcffRecalc = False # hacky
		self.lastMarkovIntMapChoice = None
		self.lastMarkovPitchChoice = None
		self.lastSegmenthLength = 1 # always start at 1
		self.incSmearAdditionalNotes = 0
		self.lastSegmentInitiated = []
		self.lastSegmentTerminated = []
		self.selectedSegs = []
		self.segmentsPerFrame = np.zeros(self.lengthInFrames, dtype=int) # notekeeper
		self.onsetsPerFrame = np.zeros(self.lengthInFrames, dtype=int) # notekeeper
		ag.allHandles.append(self) # add to all list
		self.data = {'raw': {}, 'norm': {}, 'mix-raw': {}}
	############################################################################
	def initAnal(self, descriptorsHandle):
		p.post(self.name, inc=True)
		SdifInterface.getDescriptorSegment(self.filename, self.startSec, self.endSec, descriptorsHandle.target_load, self.data['raw'], self.envelopeMask)
		for d in descriptorsHandle.target_seg:
			self.data['raw'][d.name] = [None]*self.lengthInFrames # make a tmp data track
		for d in descriptorsHandle.target_onfly:
			self.data['raw'][d.name] = metrics.analSegment(d, self, None, None, "Calc", ag)
		self.maxPower = np.max(self.data['raw']['power'])
		self.minPower = np.min(self.data['raw']['power'])
	############################################################################
	def increment(self, frameIncrement):
		if type(frameIncrement) in [int, float]: # increment with how long it takes to drop INT db, else the seg length
			thisSegMaxPower = max(self.data[ops.TARGET_INCREMENT_ENVELOPE]['power'][self.timeInFrames:self.timeInFrames+self.nextIncrementLength])
			thisMomentDbFromPeakDb = utilities.ampToDb(thisSegMaxPower/self.maxPower)
			if thisMomentDbFromPeakDb > frameIncrement:
				self.incSmearAdditionalNotes += 1
				frameIncrement = 'frame'
			else:
				frameIncrement = 'segment'
		if frameIncrement == 'frame':
			self.timeInFrames += ops.TARGET_MIN_INCREMENT
		else:
			self.timeInFrames += self.nextIncrementLength
			self.nextIncrementLength = ops.TARGET_MIN_INCREMENT
	############################################################################
	def setNextInc(self, incOpt, tgtEffDur, srcEffDur):
		if type(incOpt) in [int, float]: # added on aug 8 2011 - then just inc by an int
			self.nextIncrementLength = int(incOpt)
		elif type(incOpt) in [list, tuple]: # added on aug 8 2011 - HACKY, for random range inc
			self.nextIncrementLength = random.randint(incOpt[0], incOpt[1])
		elif incOpt == "least":
			self.nextIncrementLength = min([tgtEffDur, srcEffDur])-1
		elif incOpt == 'most':
			self.nextIncrementLength = max([tgtEffDur, srcEffDur])-1
		elif incOpt in ["sourceEffDur", "srcEffDur"]:
			self.nextIncrementLength = srcEffDur-1
		elif incOpt in ["targetEffDur", "tgtEffDur", "segment", "seg"]:
			self.nextIncrementLength = tgtEffDur-1
	############################################################################
	def getCurrentSegNumber(self, time, segs):
		currentSeg = False
		for i, seg in enumerate(segs):
			if time >= seg[0] and time <= seg[1]:
				currentSeg = i
				break
		if currentSeg == False: currentSeg = 0
		return currentSeg
	############################################################################
	def rewind(self):
		# called it tgt is used for another concate..
		self.timeInFrames = 0
		self.segmentsPerFrame = np.zeros(self.lengthInFrames, dtype=int) # notekeeper
		self.onsetsPerFrame = np.zeros(self.lengthInFrames, dtype=int) # notekeeper
		self.data['raw'] = deepcopy(self.data['orig']) # restore unsubtracted/re-analized target data....
		self.onsetInit()
		src.fileSelectedList = []
		src.nameSelectedList = []
		src.simSelectedList = []
		self.initSegMethod(descriptorData.target_seg)
		self.lastMarkovIntMapChoice = None
		self.lastMarkovPitchChoice = None
		src.timesVoiceWasPlayed = {}
		ag.selectedSegmentTimes = []
		ag.selectedSegments = [] # maintain a list of selected corpus segments
		ag.activeSegments = [] # maintain a list of currently active corpus segments
		ag.countRanOutOfSamples = 0
		self.initMixture()
	############################################################################
	def onsetInit(self):
		'''this initializes data for triggerings methods.'''
		self.triggerData = {'totalWeights': 0}
		for d in descriptorData.master_list:
			if d.origin != 'TARGET_ONSET_DESCRIPTORS': continue
			if d.weight <= 0: continue
			self.triggerData[d.name] = {}
			if self.data['raw'].has_key(d.name) == False: self.data['raw'][d.name] = metrics.analSegment(d, self, None, None, "Calc", ag)
			if ops.TARGET_ONSET_FORCE_MAX.has_key(d.name): # hacked to force max val
				self.triggerData[d.name]['initMax'] = ops.TARGET_ONSET_FORCE_MAX[d.name]
			else:
				self.triggerData[d.name]['initMax'] = max(self.data['raw'][d.name])

			self.triggerData[d.name]['initMin'] = min(self.data['raw'][d.name])
			self.triggerData['totalWeights'] += d.weight
			if d.name.find('power') != -1 or d.name.find('rms') != -1 or d.name.find('flux') != -1:
				self.triggerData[d.name]['type'] = 'power'
			else:
				self.triggerData[d.name]['type'] = 'metric'
		self.triggerScalar = ops.TARGET_ONSET_THRESHOLD
		# for simultanous selection amplitude scaling...
		self.triggerThreshNumbNotes = {}
		self.triggerThreshActiveNumbNotes = {} # a higher threshold for subtraced areas?
		for i in range(ops.MAX_SEGMENTS_PER_ONSET+1):
			self.triggerThreshNumbNotes[i] = utilities.dbToAmp(self.triggerScalar+(ops.TARGET_ONSET_THRESHOLD_SIM_PENALITY*i))*self.triggerData['totalWeights']
		# different thresh if there has been a subtraction?
		if ops.TARGET_ONSET_THRESHOLD_ACTIVE_PENALITY != None:
			self.triggerScalarActive = utilities.dbToAmp(ops.TARGET_ONSET_THRESHOLD_ACTIVE_PENALITY)
		else: self.triggerScalarActive = 1
	############################################################################
	def offsetInit(self): 
		'''this initializes data for segment offsets.'''
		# avoid 0 powers due to the fades..
		#if ops.TARGET_FADE_IN_SEC not in [0, None]:
		#	start = int(ops.TARGET_FADE_IN_SEC/ag.frameSec)
		#else:
		#if ops.TARGET_FADE_OUT_SEC not in [0, None]:
		#	end = len(self.data['raw']['power'])-int(ops.TARGET_FADE_OUT_SEC/ag.frameSec)
		#else:
		minPower = min(self.data['raw']['power'])
		if ops.TARGET_SEGMENT_LENGTH_DROP_METHOD == 'absDb':
			# X db higher than min...
			if minPower == 0:
				addToPrevent0 = 0.0001
			else:
				addToPrevent0 = 0
			self.absDbDropPower = ((minPower+addToPrevent0)*utilities.dbToAmp(ops.TARGET_OFFSET_DB_ABOVE_MIN))-addToPrevent0
			self.absDbDropPower = (0.00001*utilities.dbToAmp(ops.TARGET_OFFSET_DB_ABOVE_MIN))
#			print self.absDbDropPower
#			print self.absDbDropPower
#			print self.absDbDropPower
#			print self.data['raw']['power']
#			print self.absDbDropPower
	############################################################################
	def thresholdTest(self, time):
		'''Test the target frame TIME to see if
		it is above the triggering threshold.'''
		thisNoteTrigger = self.triggerThreshNumbNotes[self.onsetsPerFrame[time]]
		if ops.TARGET_ONSET_THRESHOLD_ACTIVE_PENALITY != None and len(ag.activeSegments) > 1:
			thisNoteTrigger *= self.triggerScalarActive
			
		triggerTotal = 0 # anything above 1 triggers
		for anal, weight in ops.TARGET_ONSET_DESCRIPTORS.iteritems(): # make sure metrics are analyzed...
			if weight <= 0: continue # skip zeros
			# use below if it is a power triggering method -- ignore negative values...
			if self.triggerData[anal]['type'] == 'power':
				#print anal, time, len(self.data[ops.TARGET_ONSET_ENVELOPE][anal])
				triggerAdd = (self.data[ops.TARGET_ONSET_ENVELOPE][anal][time])/self.triggerData[anal]['initMax']
			# else take the absolute value and scale according to 
			elif self.triggerData[anal]['type'] == 'metric':
				if self.data[ops.TARGET_ONSET_ENVELOPE][anal][self.timeInFrames] >= 0:
					triggerAdd = (self.data[ops.TARGET_ONSET_ENVELOPE][anal][time])/self.triggerData[anal]['initMax']
				else: # if negative...
					triggerAdd = (self.data[ops.TARGET_ONSET_ENVELOPE][anal][time])/self.triggerData[anal]['initMin']
			triggerTotal += triggerAdd*weight
		if triggerTotal >= thisNoteTrigger: # trigger it?
			self.lastSegmentInitiated = ['thresh', utilities.trunc(utilities.ampToDb(triggerTotal), 1), utilities.trunc(utilities.ampToDb(thisNoteTrigger), 1)]
			return True
		else:
			return False
	############################################################################
#	def NEWthresholdTest(self, time, overlapList, maxOverlaps, onsetList, minOnsets, maxOnsets):
#		if maxOverlaps != None and overlapList[time] >= maxOverlaps: return False # too many overlaps!
#		if onsetList[time] < minOnsets: return True # need another onset regardless of amp thresh
#		if onsetList[time] >= maxOnsets: return False # too many onsets!
#		return self.thresholdTest(time) # return whether it is loud enough
	############################################################################
	def segmentSuperimpositionTest(self, time, overlapList, minOverlaps, maxOverlaps, onsetList, minOnsets, maxOnsets):
		'''tests to see if a segment should be selected based
		on the number of min/max onsets per frame, and max 
		overlaps per frame'''
		# first we see if we have exceeded the maximum number of segment overlaps for this particular frame
		if maxOverlaps != None and overlapList[time] >= maxOverlaps:
			return False,'max overlapping segments' # too many overlaps!
		# next we see if we have exceeded the maximum number of onsets for this particular frame
		if onsetList[time] >= maxOnsets:
			return False,'max segments per onset' # too many onsets!
		# next we see if we have not filled the minimum number of onsets for this particular frame, regardless of amp
		if onsetList[time] < minOnsets and onsetList[time] > 0:
			return True,'minimum onsets' # need another onset regardless of amp thresh
		if minOverlaps != None and overlapList[time] < minOverlaps and overlapList[time] > 0:
			return True,'minimum overlaps' # need another overlap regardless of amp thresh
		# if passing this point, test with the onset detector
		ampThresh = self.thresholdTest(time) # whether it is loud enough
		if not ampThresh:
			return False,'below onset threshold'
		else:
			return True,'yes'
	############################################################################
	def findRawSegmentLength(self, time):
		if self.data['raw']['effDur-seg'][time] == None:
			ed = descriptorData.getDescriptorObjByName('effDur-seg')
			ag.segAnal(self, ed, time, None)
		return self.data['raw']['effDur-seg'][time]
	############################################################################
	def initSegMethod(self, descriptors): # get target triggers all at once for segment calc and normalization
		if ag.targetSegmentsFromFile:
			for fi,fiend in tgt.segFrames:
				for d in descriptorData.target_seg:
					ag.segAnal(self, d, fi, fiend)
		else:
			for fi in range(0, tgt.lengthInFrames-1):
				if self.thresholdTest(fi): # then its a hit
					segmentLength = self.findRawSegmentLength(fi)
					for d in descriptorData.target_seg:
						ag.segAnal(self, d, fi, fi+segmentLength)
	############################################################################
	def getSegs(self, path):
		segs = []
		self.segFrames = [] # an internal list of segments in frame-time
		ag.targetSegmentsFromFile = False # just to keep track
		self.segFrameOnsets = []
		if ops.TARGET_LOAD_SEGMENTS_FROM_FILE not in [False, None]:
			# load target segments from a file
			exists, path, fileType = path.test(ops.TARGET_LOAD_SEGMENTS_FROM_FILE, '/') 
			ag.targetSegmentsFromFile = True
			times = utilities.readAudacityLabelFile(path, False, False)
			for i in range(len(times)):
				# test to make sure the segments are within start and end second
				if times[i][0] < self.startSec or times[i][1] > self.endSec: continue
				segmentStart = times[i][0]-self.startSec
				segmentLengthFrames = times[i][1]-segmentStart
				segs.append((segmentStart, segmentLengthFrames))
				self.segFrames.append((int(segmentStart/ag.frameSec), int(float(times[i][1])/ag.frameSec)))
				self.segFrameOnsets.append(int(segmentStart/ag.frameSec))
			if len(segs) == 0:
				utilities.error("\n\n\nNo segments FOUND!?!?!?!\n\n")
			return segs
		# else analyze segments on the fly (the default)
		fi = 0
		while fi < self.lengthInFrames-1:
			if self.thresholdTest(fi): # then its a hit
				segmentLengthFrames = self.findRawSegmentLength(fi)
				segmentStart = fi*ag.frameSec
				segs.append((segmentStart, segmentLengthFrames*ag.frameSec, self.lastSegmentInitiated, self.lastSegmentTerminated))
				self.segFrames.append((fi, fi+segmentLengthFrames))
				fi += segmentLengthFrames
			else:
				fi += 1
		return segs
	############################################################################
	def writeSegLabelsToDisk(self, path, segs, tempo):
		scale = 60.0/float(tempo)
		segs.sort()
		labels = open(path, "w")
		for i, seg in enumerate(segs):
			writeStr = str(seg[0]*scale)+'\t'+str((seg[0]+seg[1])*scale)
			if seg != None and len(seg) > 2 and len(seg[2]) > 1:
				writeStr += "\tonset "+seg[2][1] # add segmentation reason
			if seg != None and len(seg) > 2 and len(seg[3]) > 1:
				writeStr += " offset "+seg[3][0]+' '+seg[3][1] # add segmentation reason
			labels.write(writeStr+"\n")
		labels.close()
	############################################################################
	def writeOMScore(self, path, segs):
		#frq amp start length channel
		startMSList = ""
		durMSList = ""
		midiAmpList = ""
		midiCentList = ""
		chanList = ""
		for seg in segs:
			startMSList += str(int(seg[0]*1000))+" "
			durMSList += str(int(seg[1]*1000))+" "
			midiAmpList += str(127)+" " # max amp
			midiCentList += str(6000)+" " # middle C
			chanList += str(0)+" "
		OmScoreFile = open(path, 'w')
		OmScoreFile.write("("+midiCentList+") ("+startMSList+") ("+durMSList+") ("+midiAmpList+") ("+chanList+")")
		OmScoreFile.close()
	############################################################################
	def writeSegSfsToDisk(self, segs):
		for i, seg in enumerate(segs):
			print "writing seg sf to disk:", i, '->', seg
			utilities.executeCommand(['sox', '"'+self.filename,os.path.join(ops.SCRIPT_PATH, "output/segs", 'tgt'+str(i)+'.aiff')+'"', 'trim', str(seg[0]), str(seg[1])])
	############################################################################
	def plot(self, normalize=True, dataNames=['centroid']):
		normalize = False
		plotFileName = os.path.join(ops.SCRIPT_PATH, "output/segs", 'plot.png')
		dataArrays = []
		for d in dataNames: # normalize
			listy = np.array(self.data['raw'][d])
			if normalize:
				miny, maxy = min(self.data['raw'][d]), max(self.data['raw'][d])
				listy = (listy-miny)/(maxy-miny)
			dataArrays.append(listy)
		plot(plotFileName, dataArrays, dataNames, segs=self.segFrames, height=3.5)
		#waterfallPlot(os.path.join(ops.SCRIPT_PATH, "output/segs", 'mels'+str(self.timeInFrames)+'.png'), self.data['raw']['melamps'])
	############################################################################
#	def initBash(self, anal, dataType):
#		if not hasattr(self, 'bashy'): self.bashy = {}
#		if not self.bashy.has_key(anal):
#			pdf, bin_edges = np.utilities.histogram(self.data[dataType][anal], bins=len(self.data[dataType][anal]), normed=False, new=True)
#			cdf = np.cumsum(pdf, dtype=float)
#			self.bashy[anal] = {'cdf': [], 'min':min(self.data[dataType][anal]), 'max':max(self.data[dataType][anal]), 'len': len(self.data[dataType][anal])-1}
#			for i, val in enumerate(cdf):
#				self.bashy[anal]['cdf'].append((val-min(cdf))/float(max(cdf)))
###########################################################################
#	def bash(self, anal, dataType, start, end):
#		raw_as_bins = []
#		for i in self.data[dataType][anal][start:end]:
#			zeroToOne = (i-self.bashy[anal]['min'])/(self.bashy[anal]['max']-self.bashy[anal]['min'])
#			zeroToOne = utilities.limit(zeroToOne, 0, 1) # clip in case of new values from recalc
#			raw_as_bins.append(int(zeroToOne*self.bashy[anal]['len']))
#		output = []
#		for i in raw_as_bins:
#			output.append(self.bashy[anal]['cdf'][i]) # from 0->1
#		for i, val in enumerate(output): # un normalize...
#			output[i] = (val*(self.bashy[anal]['max']-self.bashy[anal]['min']))+self.bashy[anal]['min']
#		return output
	############################################################################
	def featureModification(self, start, end, dataSlot):
		for [d, modType, value] in descriptorData.target_feature_mod:
			self.calcFeatureWarp(d.name, value, modType, dataSlot, start, end)
	############################################################################
#	def featureBashing(self, item, dataIn, dataOut, start, end):
#		if item in ops.TARGET_DESCRIPTORS_BASH or item == 'all':
#			if item == 'all': anals = analMetrics.norm
#			else: anals = [item]
#			for anal in anals:
#				if analMetrics.metric[anal]['segmented']: continue # skip seg methods here...
#				print "BASHING TARGET", anal
#				self.initBash(anal, dataIn)
#				output = self.bash(anal, dataIn, start, end)
#				#plotFileName = os.path.join(ops.SCRIPT_PATH, ops.PLOT_PATH, 'bash.png')
#				#plot(plotFileName, [self.data['raw']['centroid'], output], ['centroid', 'bashed'], segs=self.segFrames, height=5)
#				self.data[dataOut][anal][start:end] = output
	############################################################################
	def calcFeatureWarp(self, metric, value, modType, dataSlot, start, end):
		# called by featureModification for each anal...
		value = float(value)
		if start == None and end == None: # this is an init pass...
			#print ""
			#print "Modifying Target post-normalization -> "+metric+modType+str(value)
			#print "\twas:", self.data['norm'][metric][0:5]
			for idx, val in enumerate(self.data[dataSlot][metric]):
				if val == None: continue # skip Nones possible with seg anals
				calcRollOffDataPoint(metric, dataSlot, idx, val, value, modType)
			#print "\tnow:", self.data['norm'][metric][0:5]
		else: # only do a slice...
			for idx, val in enumerate(self.data[dataSlot][metric][start:end]):
				if val == None: continue # skip Nones possible with seg anals
				calcRollOffDataPoint(metric, dataSlot, start+idx, val, value, modType)
	############################################################################
	def onFlySegmentCalc(self, d, start, end, dataType):
		if type(d) == str: d = descriptorData.getDescriptorObjByName(d)
		# returns segment calculation
		if dataType == 'norm':
			if self.data['norm'][d.name][start] == None: # if segment calc for this target frame has not been computed...
				ag.segAnal(self, d, start, end)
				ag.normalize([self], [d.name], 'raw', 'norm', ops.NORMALIZE_METHOD, ag.targetNormCoeffs, start, start+1)
				self.featureModification(start, start+1, 'norm')
			return self.data['norm'][d.name][start]
		elif dataType in ['raw']:
			if self.data['raw'][d.name][start] == None: # if segment calc for this target frame has not been computed...
				ag.segAnal(self, d, start, end)
				self.featureModification(start, start+1, 'raw')
			return self.data['raw'][d.name][start]
	############################################################################
	def leastSqrsAmp(self, sampleHandle, minGain, maxGain, subtractTime, subtractDuration):
		minFrames = min([subtractDuration, self.lengthInFrames-self.timeInFrames, sampleHandle.lengthInFrames])
		tgtFrames = self.data['raw']['power'][subtractTime:subtractTime+minFrames]
		srcFrames = sampleHandle.data['raw']['power'][:minFrames]#*utilties.dbToAmp(sampleHandle.scaleDb)
		try:
			leastSqrWholeLine = (np.linalg.lstsq(np.vstack([srcFrames]).T, np.vstack([tgtFrames]).T)[0][0][0])
		except np.linalg.linalg.LinAlgError: # in case of incompatible dimensions
			leastSqrWholeLine = 0
			pass
		sourceAmpScale = utilities.limit(leastSqrWholeLine, utilities.dbToAmp(minGain), utilities.dbToAmp(maxGain))
		return sourceAmpScale, 100+utilities.ampToDb(sourceAmpScale)
	############################################################################
	def peakAmpCorrect(self, sampleHandle, minGain, maxGain, subtractTime, subtractDuration):
		minFrames = min([subtractDuration, self.lengthInFrames-self.timeInFrames, sampleHandle.lengthInFrames])
		tgtPower = self.onFlySegmentCalc(ops.POST_SELECTION_AMP_PEAK_AMP_METRIC, subtractTime, subtractTime+subtractDuration, "raw")
		srcPower = sampleHandle.data['raw'][ops.POST_SELECTION_AMP_PEAK_AMP_METRIC][0]
		scale = tgtPower/srcPower
		return scale, 100+utilities.ampToDb(scale)
	############################################################################
	def averagedAmpCorrect(self, sampleHandle, minGain, maxGain, subtractTime, subtractDuration):
		'''averages the scalar difference of the loudest 50% of target frames!'''
		percentage = 50.
		minFrames = min([subtractDuration, self.lengthInFrames-self.timeInFrames, sampleHandle.lengthInFrames])
		evalFrame = int((percentage/100.)*minFrames)
		tgtFrames = self.data['raw']['power'][subtractTime:subtractTime+minFrames]
		srcFrames = sampleHandle.data['raw']['power'][:minFrames]
		tmp = []
		for fidx, frame in enumerate(tgtFrames):
			tmp.append((frame, fidx))
		tmp.sort()
		tmp.reverse()
		f = 0
		avgs = []
		while f <= evalFrame:
			tFrame, fidx = tmp[f]
			avgs.append(tFrame/srcFrames[fidx])
			f += 1
		scale = np.average(avgs)
		return scale, 100+utilities.ampToDb(scale)
	############################################################################
	def initMixture(self):
		self.data['mix-raw'] = {}
		for d in descriptorData.target_descriptor_mixture:
			if d.seg:
				self.data['mix-raw'][d.name] = {}
			else:
				self.data['mix-raw'][d.name] = np.zeros(self.lengthInFrames)
	############################################################################
	def mixSelectedSamplesDescriptors(self, sampleHandle, sourceAmpScale, mixTime, mixDuration):
		#print "\nDESCRIPTOR MIX", sampleHandle.name, "\ttarget time:", mixTime, "\tmix duration:", mixDuration
		# catch if target requested too far:
		if mixTime+mixDuration > self.lengthInFrames: mixDuration = self.lengthInFrames-mixTime # if too long for tgt
		if mixDuration > sampleHandle.lengthInFrames: mixDuration = sampleHandle.lengthInFrames # if too long for corpus item
		for d in descriptorData.target_descriptor_mixture:
			if not d.seg:
				self.data['mix-raw'][d.name][mixTime:mixTime+mixDuration] =  metrics.mixTwoTimeVaryingFeatures(d.name, mixTime, mixDuration, self.data['mix-raw'], sampleHandle.data['raw'])
			else:
				ag.clearSegDict(self.data['mix-raw'][d.name], mixTime, mixDuration)
	############################################################################
	def subtractPowerForOnsetEnvelope(self, sampleHandle, sourceAmpScale, start, length):
		length = min(self.lengthInFrames-start, length, sampleHandle.lengthInFrames)
		samplePowers = sampleHandle.data['raw']['power'][0:length]
		rawSubtraction = self.data['raw']['power'][start:start+length]-(samplePowers*ops.SCALE_POWER_SUBTRACTION*sourceAmpScale)
		# clip it so its above zero
		self.data['raw']['power'][start:start+length] = np.clip(rawSubtraction, 0, sys.maxint)
		# recalculate onset envelope
		for d in descriptorData.target_descriptor_subtract:
			if d.name == 'power': continue # its already been done
			self.data['raw'][d.name][start:start+length] = metrics.analSegment(d, self, start, start+length, "Recalc", ag)
		# erase power-segs..
		for time in range(start,start+length):
			tgt.data['raw']['power-seg'][time] = None
			if tgt.data['norm'].has_key('power-seg'): tgt.data['norm']['power-seg'][time] = None
############################################################################
	def initSegmentList(self, initialSegmentList): # timeList
		orderedSegList = []
		for seg in initialSegmentList:
			power = self.onFlySegmentCalc('power-seg', seg[0], seg[1], 'raw')
			peak = metrics.peakTimeSeg(self.data[ops.CONCATENATE_PEAK_ALIGN_ENVELOPE]['power'][seg[0]:seg[1]])[0]
			orderedSegList.append([power, seg[0], seg[1], peak])
			# segmentList = [power, start, end, peakTime]
		orderedSegList.sort()
		orderedSegList.reverse()
		self.lengthOriginalSegs = len(orderedSegList)
		return orderedSegList
############################################################################
	def updateSegmentList(self, segmentList, start, end):
		# this takes segment list and
		#		removes first item
		#		subtracts power of selected corpus segment
		#		reevaluates power of first segment region
		#		adds new segments if needed
		newSegments = []
		fi = start
		if not ag.targetSegmentsFromFile:
			while fi < end:
				if self.thresholdTest(fi): # then its a hit
					segmentLengthFrames = self.findRawSegmentLength(fi)
					power = self.onFlySegmentCalc('power-seg', fi, fi+segmentLengthFrames, 'raw')
					peak = metrics.peakTimeSeg(self.data[ops.CONCATENATE_PEAK_ALIGN_ENVELOPE]['power'][fi:fi+segmentLengthFrames])[0]
					newSegments.append([power, fi, fi+segmentLengthFrames, peak])
					fi += segmentLengthFrames
				else:
					fi += 1
		else:
			# just keep the same segment length ...
			power = self.onFlySegmentCalc('power-seg', start, end, 'raw')
			peak = metrics.peakTimeSeg(self.data[ops.CONCATENATE_PEAK_ALIGN_ENVELOPE]['power'][start:end])[0]
			newSegments.append([power, start, end, peak])
#		print segmentList[0], "became:", newSegments
#		print ops.SCALE_POWER_SUBTRACTION
#		print segmentList[0], "became:", newSegments
#		print segmentList[0], "became:", newSegments
		del segmentList[0]
		if len(newSegments) == 0: self.lengthOriginalSegs -= 1
		newSegments.extend(segmentList)
		#segmentList.append(newSegments)
		#segmentList.sort()
		#segmentList.reverse()
		return newSegments

		
################################################################################
################################################################################
################################################################################
################################################################################
################################################################################
class addSfSegment:
	def __init__(self, segDict):
		self.typed = 'source'
		exists, self.filename, fileType = path.test(segDict['file'], ops.SOURCE_PATH) # user path, relative path
		if not exists:
			utilities.error('corpus', "file "+self.filename+" not found!")
		self.sfdur, self.sfchannels = SdifInterface.checkSdifResource( self, self.filename )
		self.name = self.filename.split('/')[-1] # short name for printing
		for key, val in segDict.iteritems():
			setattr(self, key, val) # turn dict items into attributes..
		ag.segTimesToWindowedLength(self)
		self.getOnsetAndOffsetTimes(self.onsetLen, self.offsetLen)
		self.data = {'raw': {}, 'norm': {}}
		self.passedLimitTest = True
	############################################################################
	def initAnal(self, descriptorsHandle):
		p.post(self.name, inc=True)
		SdifInterface.getDescriptorSegment(self.filename, self.startSec, self.endSec, descriptorsHandle.corpus_load, self.data['raw'], self.envelopeMask)
		for d in descriptorsHandle.corpus_onfly:
			self.data['raw'][d.name] = metrics.analSegment(d, self, None, None, "Calc", ag)
	############################################################################
	def getOnsetAndOffsetTimes(self, onset, offset):
		noteDuration = self.endSec-self.startSec
		cushion = 0.01
		if onset >= 1: self.onsetTime = noteDuration*(onset/100.)
		else: self.onsetTime = onset
		if offset >= 1: self.offsetTime = noteDuration*(offset/100.)
		else: self.offsetTime = offset
		if (self.onsetTime+self.offsetTime+cushion) > noteDuration:
			self.offsetTime = noteDuration-self.onsetTime-cushion
		# if SdifLinkage will factor in the envelope applied to this segment..
		onsetFrames = int(round(self.onsetTime*(1/ag.frameSec)))
		offsetFrames = int(round(self.offsetTime*(1/ag.frameSec)))

		if onsetFrames <= 1 and offsetFrames <= 1:
			self.envelopeMask = None
		else:
			self.envelopeMask = np.ones(self.lengthInFrames)
			if onsetFrames > 0:
				self.envelopeMask[:onsetFrames] = np.linspace(1/float(onsetFrames), 1, num=onsetFrames)
			if offsetFrames > 0:
				self.envelopeMask[offsetFrames*-1:] = np.linspace(1, 1/float(offsetFrames), num=offsetFrames)
	############################################################################
	def testLimits(self, limitData, maxMin, orderings):
		# the goal is to set self.passedLimitTest to False if if descriptors 
		# fail one of the local or global limitations set by the user
		for key, val in self.limit.iteritems():
			# get min and max scoped for either local entry of global corpus
			if self.limit[key]['scope'] == 'global':
				minMaxData = maxMin['global'][key]
				handleOrder = orderings['global'][key]
			elif self.limit[key]['scope'] == 'local':
				minMaxData = maxMin[self.name][key]
				handleOrder = orderings[self.name][key]
			thisHandlePercentage = handleOrder.index(self)/float(len(handleOrder))
			# do tests
			for r_type, r_value in val.iteritems(): # for each limit in the dict
				if r_type == "scope": continue # internal
				if r_type == "low":
					if limitData[self][key] < r_value: self.passedLimitTest = False
				elif r_type == "high":
					if limitData[self][key] > r_value: self.passedLimitTest = False
				elif r_type == "lowPer":
					if thisHandlePercentage < (r_value/100.): self.passedLimitTest = False
				elif r_type == "highPer":
					if thisHandlePercentage > (r_value/100.): self.passedLimitTest = False
				else:
					print "\n\tERROR; I don't know the limit key "+r_type+"!\n"
					utilities.exit(1)
			if self.passedLimitTest:
				#print "PASSED", val, ":", limitData[self]
				pass
	############################################################################
	def loadSegmentPostAnal(self):
		if not self.passedLimitTest:
			p.post("Skipping", '...', 3)
			del self.data
			return
		p.post("Loading", '...', 3)
		ag.allHandles.append(self) # add to all list
		ag.sourceHandles.append(self) # add to all list
		src.pool[self.voiceID][self] = 0 # add to mutable source pool
		ag.srcMaxPowers.append(max(self.data['raw']['power']))
		ag.corpusLengthInSeconds += self.endSec-self.startSec
			



########################################################
class otherSavedData():
	def __init__(self, fullPath, name):
		# using json
		self.fullPath = fullPath
		self.handle = None
		self.needToWrite = False
		self.triedToRead = False
	####################################################
	def hasData(self, hash):
		# open the file and read if needed...
		if not self.triedToRead:
			if os.path.isfile(self.fullPath): 
				print "Reading json archive", self.fullPath
				self.handle = open(self.fullPath, 'r')
				self.data = json.load(self.handle)
				self.handle.close()
				self.handle = None
			else:
				self.data = {}
			self.triedToRead = True
		return self.data.has_key(hash)
	####################################################
	def getData(self, hash): return self.data[hash]
	####################################################
	def commitData(self, hash, data):
		self.data[hash] = data
		#print self.data[hash]
		self.needToWrite = True
	####################################################
	def close(self):
		if self.needToWrite:
			self.handle = open(self.fullPath, 'w')
			print "writing", self.handle
			json.dump(self.data, self.handle)
	


################################################################################
################################################################################
################################################################################
################################################################################
################################################################################
class outputData:
	def __init__(self, scoreName, csoundBool, qlistName, qlistBool, tgt, csoundFileName, OmFileName, OmBool, MidifileName, midiFileBool):
		#self.scoreName = scoreName
		self.totalOutputNotes = 0
		self.scoreNotes = []
		self.scoreName = path.locate(scoreName)
		self.lastNoteTime = 0.0
		self.lastNoteEndTime = 0.0 #not needed anymore
		if qlistBool: self.qlistHandle = open(qlistName, 'w')
		if OmBool: self.OmScoreFile = open(OmFileName, 'w')
		self.midiFileName = MidifileName
		self.csoundFileName = csoundFileName
		self.scoreMixHandle = open(os.path.join(ops.TMP_PATH, 'agMix.sco'), 'w')
		self.tgtScoreSegmentHandle = open(os.path.join(ops.TMP_PATH, 'tgtSegs.sco'), 'w')
		self.qlistNotes = []
		self.OmNotes = []
		self.noteHandles = []
		self.handleToNote = {}
		self.minMaxSearch = []
		for seg in tgt.segs:
			self.tgtScoreSegmentHandle.write('\t'.join(['i1', str(seg[0]), str(seg[1]), '"'+path.locate(tgt.filename)+'"', str(seg[0]), '0.5\n']))
		tgtDur = tgt.endSec-tgt.startSec
		#CSOUND_OUTPUT_MIX_PAN = 'middle' # 'middle' || 'leftRight'
		if ops.CSOUND_OUTPUT_MIX_PAN == 'middle': pan = ('1', '1')
		elif ops.CSOUND_OUTPUT_MIX_PAN == 'leftRight': pan = ('0', '1')
		self.scoreMixHandle.write('\t'.join(['i1', '0', str(tgtDur), '"'+path.locate(tgt.filename)+'"', str(tgt.startSec), pan[0], '\n']))
		self.scoreMixHandle.write('\t'.join(['i1', '0', str(tgtDur+3), '"'+csoundFileName+'"', '0', pan[1], '\n']))
		self.outputLabelData = []
	############################################################################
	def changeOutputScoreName(name):
		self.scoreName = path.locate(name)
	############################################################################
	def pstr(self, val):
		return utilities.trunc(val, ops.OUTPUT_FLOAT_PRECISION) # N decimal places for floats.
	############################################################################
	def addEvent(self, handle, timeSec, durSec, targetDuration, outputDb, trans, segCntId):
		peakTimeSec = handle.data['raw']['peakTime-seg'][0]*ag.frameSec
		effDurSec = handle.data['raw']['effDur-seg'][0]*ag.frameSec
		rawRms = utilities.dbToAmp(utilities.ampToDb(handle.data['raw']['power-seg'][0]))
		outputDb += ops.OUTPUT_GAIN+handle.scaleDb
		if ops.CSOUND_MAKE_SCORE:
			outputFiles.addScoreNote("i1", timeSec, durSec, outputDb, handle, trans, handle.onsetTime, handle.offsetTime, peakTimeSec, effDurSec, handle.voiceID, len(ag.activeSegments), targetDuration/float(durSec), segCntId) # add note to csound sco data
		if ops.MAKE_QLIST:
			self.addQlistNote(handle, timeSec, trans, outputDb, effDurSec, handle.onsetTime, handle.offsetTime)

		if ops.SAVE_OUTPUT_SEGMENT_LABELS:
			self.outputLabelData.append([self.pstr(timeSec), self.pstr(timeSec+(handle.endSec-handle.startSec)), handle.name+'@'+self.pstr(handle.startSec)+'-'+self.pstr(handle.endSec), ' '.join(handle.extraMappedData).replace('\n',''), '\n'])
	############################################################################
	def addComment(self, listy):
		if ops.CSOUND_MAKE_SCORE: self.scoreNotes.append(listy)
	############################################################################
	def addScoreNote(self, instru, startTime, absDur, amp, handle, semitones, onsetTime, offsetTime, peakTimeSec, effDur, voiceID, simSelectVoice, stretchCoeff, segCntIdx):
		thisTime = startTime*(60.0/float(ops.TEMPO_CHANGE))
		# get output duration
#		if ops.CSOUND_SCORE_DURATION_TYPE == 'dur': 
#			absDur = handle.endSec-handle.startSec
#		elif ops.CSOUND_SCORE_DURATION_TYPE == 'effDur':
#			absDur = effDur
#		elif ops.CSOUND_SCORE_DURATION_TYPE == 'matchDur':
#			absDur = simCalcDuration*ag.frameSec
		#absDur = handle.endSec-handle.startSec
		semitones = utilities.quantize(semitones, ops.TRANSPOSE_QUANTIZE)
		trans = 2**(semitones/12.0)
		absDur = absDur*(1/trans)
		if ops.LIMIT_SEGMENT_DURATIONS != None:
			noteDuration = min(absDur, ops.LIMIT_SEGMENT_DURATIONS)
		else:
			noteDuration = absDur
				
		self.handleToNote[handle] = [instru, self.pstr(thisTime), self.pstr(noteDuration), self.pstr(amp), '"'+os.path.abspath(handle.filename)+'"', self.pstr(handle.startSec), self.pstr(trans), self.pstr(utilities.ampToDb(handle.data['raw']['power-seg'][0])), self.pstr(peakTimeSec), self.pstr(effDur), self.pstr(onsetTime), self.pstr(offsetTime), str(int(voiceID)), str(int(simSelectVoice)), str(stretchCoeff), str(segCntIdx)]

		self.scoreNotes.append([instru, self.pstr(thisTime), self.pstr(noteDuration), self.pstr(amp), '"'+os.path.abspath(handle.filename)+'"', self.pstr(handle.startSec), self.pstr(trans), self.pstr(utilities.ampToDb(handle.data['raw']['power-seg'][0])), self.pstr(peakTimeSec), self.pstr(effDur), self.pstr(onsetTime), self.pstr(offsetTime), str(int(voiceID)), str(int(simSelectVoice)), str(stretchCoeff), str(segCntIdx)])
		if handle.data['raw']['MIDIPitch-seg'][0] != -1:
			thisFrq = handle.data['raw']['MIDIPitch-seg'][0]+semitones
		elif handle.mappedPitch != -1:
#			print handle.name, handle.mappedPitch, handle.extraMappedData
#			print handle.name, handle.mappedPitch, handle.extraMappedData
#			print handle.name, handle.mappedPitch, handle.extraMappedData
			thisFrq = float(handle.mappedPitch)
		else:
			thisFrq = 60 # default frq of 60 for MIDI if cannot find pitch
		self.OmNotes.append([thisTime, noteDuration, amp+ops.OUTPUT_GAIN, handle.name, thisFrq, utilities.ampToDb(handle.data['raw']['power-seg'][0]), peakTimeSec, effDur, onsetTime, offsetTime, handle, stretchCoeff])
		self.lastNoteTime = startTime
		self.lastNoteEndTime = startTime+absDur
		self.totalOutputNotes += 1 # increment
		src.fileSelectedList.append(handle.name)
		src.nameSelectedList.append(handle.name)
		self.noteHandles.append([handle, thisTime, amp+ops.OUTPUT_GAIN, semitones])
	############################################################################
	def writeOMScoreNote(self):
		WRITE_REAL_DURATIONS = True
		#frq amp start length channel
		startMSList = ""
		durMSList = ""
		midiAmpList = ""
		midiCentList = ""
		chanList = ""
		self.OmNotes.sort() # fuck this
		notesMaybeQuant = utilities.quantizeNoteSequence(self.OmNotes, ops.RHYTHM_QUANTIZE)
		for (thisTime, noteDuration, amp, name, MIDIpitch, rmsPeak, peakTimeSec, effDur, onsetTime, offsetTime, handle, stretchCoeff) in notesMaybeQuant:
			startMSList += str(int(thisTime*1000))+" "
			if WRITE_REAL_DURATIONS:
				durMSList += str(int(effDur*1000))+" "
			else:
				durMSList += str(10)+" "
			thisAmp =  utilities.limit((amp-100+(rmsPeak)), -40, 0)
			
			midiAmpList += str((thisAmp*3.175)+127)+" "
			midiCentList += str((MIDIpitch)*100)+" "
			chanList += str(0)+" "
		self.OmScoreFile.write("("+midiCentList+") ("+startMSList+") ("+durMSList+") ("+midiAmpList+") ("+chanList+")")
		self.OmScoreFile.close()
	############################################################################
	def writeLispEvents(self, filename):
		WRITE_REAL_DURATIONS = False # if not, write eff durations
		self.OmNotes.sort()
		handle = open(filename, 'w')
		handle.write('( ')
		for idx, (thisTime, noteDuration, amp, name, MIDIpitch, rmsPeak, peakTimeSec, effDur, onsetTime, offsetTime, handle, stretchCoeff) in enumerate(self.OmNotes):
			##print (thisTime, noteDuration, amp, name, MIDIpitch, rmsPeak, peakTimeSec, effDur, onsetTime, offsetTime)
			handle.write( '('+self.pstr(thisTime)+' '+self.pstr(effDur)+' '+self.pstr(MIDIpitch)+' "'+name+'" '+self.pstr(rmsPeak)+') ')
		handle.write(')')
		handle.close()
	############################################################################
	def writeBachScoreNotes(self, fullPath, sourceHandles, csoundOutputFile):
		instuNames = []
		clefs = []
		keys = []
		midiChns = []
		polyphonys = []
		noteOrganizer = {} # a dict of instrNames with a list of notes grouped by onset time
		# get the order of parts
		partOrder = []
		activeInstrNames = []
		for start, end, handle, amp in ag.selectedSegments:
			if handle.MWinstrName not in activeInstrNames: activeInstrNames.append(handle.MWinstrName)

		for instu, dicty in ops.MW_DEFINE_INSTRUMENTS:
			if instu == 'noName': continue
			if instu not in activeInstrNames: continue # only add instruments that get played
			partOrder.append(instu)
			noteOrganizer[instu] = []
		#'MWtext': 'texty'
		# get part metadata
		outlist = []
		for ins in partOrder:
			clefs.append(str(src.MWData[ins]['clef']))
			keys.append(str(src.MWData[ins]['key']))
			midiChns.append(str(src.MWData[ins]['midiChn']))
			polyphonys.append(src.MWData[ins]['polyphony'])
		commandList = []
		# write header
		commandList.append(['reset', 'bang'])
		commandList.append(['target', path.locate(tgt.filename), str(tgt.startSec*1000), str(tgt.endSec*1000)])
		#commandList.append(['rendered', csoundOutputFile])
		commandList.append(['both', 'clefs', ' '.join(clefs)])
		commandList.append(['both', 'keys', ' '.join(keys)])
		commandList.append(['both', 'voicespacing', '6', '20 '*len(partOrder), '0'])
		commandList.append(['both', '(slotinfo (3 (name technique) (type text)) (4 (name sfinfo) (type text)) (5 ( name amplitudeEnvelope ) ( type function ) ( range 0. 2. ) ( representation ) ( ysnap ( ) ) ( domain 0. 1. ) ( domainslope 0. ) ( width temporal ) ( singleslotfortiednotes 1 ) ))'])

		# note loop
		self.OmNotes.sort()
		for idx, (thisTime, noteDuration, amp, name, MIDIpitch, rmsPeak, peakTimeSec, effDur, onsetTime, offsetTime, handle, stretchCoeff) in enumerate(self.OmNotes):
			if not noteOrganizer.has_key(handle.MWinstrName): continue
			ampShape = list((handle.data['raw']['power'][:]-min(handle.data['raw']['power']))/max(handle.data['raw']['power']))
			ampShapeStr = ''
			timeReduceAmpSlope = 3
			if len(ampShape) > 0:
				for aidx, amp in enumerate(ampShape[::timeReduceAmpSlope]):
					timeZeroToOne = utilities.trunc(aidx/float(len(ampShape)/timeReduceAmpSlope), 2)
					ampShapeStr += '('+timeZeroToOne+' '+utilities.trunc(amp, 2)+' 1) '
				ampShapeStr += ' (1. 0. 1)' # cap the end
			noteDuration = effDur
			if src.MWData[handle.MWinstrName]['shortestDur']: noteDuration = 0.1
			noteOrganizer[handle.MWinstrName].append([thisTime, str(MIDIpitch*100), noteDuration, str(100), handle.MWtext, handle.name+' '+self.pstr(handle.startSec)+'-'+self.pstr(handle.endSec), ampShapeStr])
		##################
		# filter notes to make continuous parts - limiting the polyphony (overlap) of each instrument
		for iidx, ins in enumerate(partOrder):
			noteOrganizer[ins].sort()
			overlapCnt = 1
			lastNoteEnded = -1
			for nidx, note in enumerate(noteOrganizer[ins]):
				if polyphonys[iidx] > 1: continue
				overlappage = lastNoteEnded-note[0]
				#print "#"+str(nidx), "start", note[0], "end", note[0]+note[2], "lastEnd", lastNoteEnded, overlappage # a little extra
				if overlappage > 0: 
					# then modify the duration of the last note
					noteOrganizer[ins][nidx-1][2] -= overlappage
				lastNoteEnded = note[0]+note[2]
		##################
		# change time to MS - both start and summed duration
		for iidx, ins in enumerate(partOrder):
			for nidx, note in enumerate(noteOrganizer[ins]):
				noteOrganizer[ins][nidx][2] = str( int((note[0]+note[2]*1000)) )
				noteOrganizer[ins][nidx][0] = str( int(note[0]*1000) )
		##################
		# write bach text
		for voiceNumb, partName in enumerate(partOrder):
			for note in noteOrganizer[partName]:
				commandList.append(['roll', 'addchord', str(voiceNumb+1), '(', note[0], '(', ' '.join(note[1:4]), '( slots ( 3 "'+note[4]+'" ) ( 4 "'+note[5]+'" ) ( 5 '+note[6]+' ) ) ) )'])
		# quantify commands
		meterStr = ("((("+ops.BACH_METER.replace('/',' ')+") ("+str(ops.BACH_TEMPO)+"))) ")*len(partOrder)
		commandList.append(['meters', meterStr])
		# write file
		commands = open(fullPath, 'w')
		for idx, command in enumerate(commandList):
			commands.write( str(idx)+', '+' '.join(command)+';\n')
		commands.close()
	############################################################################
	def writePWGLScoreNote(self, fullPath):
		WRITE_SIMPLE_RESTS = False
		dynamics = ['pp', 'p', 'mp', 'mf', 'f', 'ff']
		simple = open('/Users/ben/Documents/audioGuide/output/simpleOnsets.lisp', 'w')
		enp = open(fullPath, 'w')
		enp.write('(((\n')
		simple.write('(')
		enpOut = []
		firstStart = self.OmNotes[0][0]
		self.OmNotes.sort()
		for idx, (thisTime, noteDuration, amp, name, MIDIpitch, rmsPeak, peakTimeSec, effDur, onsetTime, offsetTime, handle, stretchCoeff) in enumerate(self.OmNotes):
			amp = utilities.limit(utilities.dbToAmp(amp-110+rmsPeak), 0, 1)
			amp = int(amp*(len(dynamics)-1))
			enpOut.append([float(thisTime-firstStart), str(MIDIpitch), float(effDur), dynamics[amp]])
		enpOutRests = []
		lastStart = -1
		lastEnd = 0.
		notes = []
		durs = []
		# add rests
		for idx, (start, midi, dur, dyn) in enumerate(enpOut):
			if WRITE_SIMPLE_RESTS and start-lastEnd > 0.4:
				# write the next note!
				simple.write(self.pstr((start-lastEnd)*-1)+" ")
				#print "WRITE",  start, dur, lastEnd, start-lastEnd
			enpOutRests.append('( '+self.pstr(start)+' :notes ('+midi+') :duration '+self.pstr(dur)+' )\n')
			simple.write(self.pstr(start)+" ") # just a list of start times
			lastEnd = start+dur # for next possible rest..
			
		for note in enpOutRests: # write the list file.
			enp.write(note)
		enp.write(')))\n')
		simple.write(')\n')
		enp.close()
		simple.close()
	############################################################################
	def addQlistNote(self, chosenSample, startTime, trans, outputDb, noteDuration, onsetTime, offsetTime):
		thisNoteStart = str(int(chosenSample.startSec*1000))
		thisNoteDur = str(int((chosenSample.endSec)*1000)-int(chosenSample.startSec*1000))
		self.qlistNotes.append([startTime, 'play', chosenSample.filename, thisNoteStart, thisNoteDur, self.pstr(outputDb), self.pstr(trans), self.pstr(int(onsetTime*1000)), self.pstr(int(offsetTime*1000))])
	############################################################################
	def writeQlistNotes(self):
		self.qlistNotes.sort()
		lastNoteTime = 0
		for note in self.qlistNotes:
			delayMS = int((note[0]-lastNoteTime)*1000)
			lastNoteTime = note[0]
			note[0] = str(delayMS)
		outputListOfLists = utilities.printableList(self.qlistNotes)
		for note in outputListOfLists:
			self.qlistHandle.write(note)
		self.qlistHandle.close()
	############################################################################
	def writeScoreNotes(self):
		if ops.ALIGN_NOTE_PEAKS:
			for lidx in range(len(self.scoreNotes)):
				if self.scoreNotes[lidx][0][0] == ';': continue # only a comment
				#print "note items:", self.scoreNotes[lidx]
				newTime = max(0, float(self.scoreNotes[lidx][1])-float(self.scoreNotes[lidx][8]))
				self.scoreNotes[lidx][1] = str(newTime)

		if not ops.STRETCH_NOTE_DURATIONS_TO_MATCH_TARGET:
			for lidx in range(len(self.scoreNotes)):
				if not self.scoreNotes[lidx][0] == "i1": continue
				self.scoreNotes[lidx][14] = "1" # bash all stretch factors to 1

		if ops.RHYTHM_QUANTIZE not in [None, 0]: # then quantize the score
			#self.scoreNotes.sort(key=lambda x: x[1])
			notePeaks = []
			for line in self.scoreNotes:
				if line[0][0] == ";": continue # just a comment
				notePeaks.append(float(line[1])+float(line[8]))
			quantizes = sliceListIntoEvenlySpacedLists(notePeaks, ops.RHYTHM_QUANTIZE, 0, max(notePeaks)+1)
			# do histogram to find most common attack point for each window...
			for idx, noteList in enumerate(quantizes): 
				if len(quantizes[idx]) > 0: quantizes[idx] = utilities.histogram(noteList)[0][1] # only if there are notes

			if ops.RHYTHM_QUANTIZE_METHOD == 'aggregate': # align notes in each window to most common attack point...
				for line in self.scoreNotes:
					if line[0][0] == ";": continue # just a comment
					idx = int((float(line[1])+float(line[8]))/ops.RHYTHM_QUANTIZE)
					line[1] = str(quantizes[idx]) # change start time of this note

#			elif ops.RHYTHM_QUANTIZE_METHOD == 'strip': # remove notes not peaking at the most common peak time...
#				for i, line in enumerate(self.scoreNotes):
#					idx = int(notePeaks[i]/ops.RHYTHM_QUANTIZE)
#					peakTime = quantizes[idx]
#					print notePeakTime, peakTime
#					if notePeakTime == peakTime:	
#						self.writeSco('\t'.join(line))
		# sort them in time
		self.scoreNotes.sort(key=lambda x: float(x[1]))
		self.scoreHandle = open(self.scoreName, 'w') # open handle
		self.scoreHandle.write(';'+tgt.filename+' '+self.pstr(tgt.startSec)+' to '+self.pstr(tgt.endSec)+'\n') # write filename to score
		self.scoreHandle.write(';'+"\t".join(['ins', 'start', 'dur', 'amp/100', 'filename', 'start-time', 'trans', 'raw-rms', 'peak-time', 'effDur', 'attack-time', 'decay-time', 'corpusID', 'simSelects', 'stretchFactor', 'tgtSegIdx'])+'\n') # write comment to score
		outputListOfLists = utilities.printableList(self.scoreNotes)
		for strline in outputListOfLists:
			self.scoreHandle.write(strline)
		self.scoreHandle.close() # close handle
		ag.scoresToRender.append(csoundRenderObj(self.scoreName, path.locate(ops.CSOUND_OUTPUT_FILE), outputFiles.totalOutputNotes))
		
		# creates dicts with voicing ID as key and a list of line numbers as val
		used = {'overlapId': {}, 'corpusId': {}}
		for nidx, noteLine in enumerate(self.scoreNotes):
			if noteLine[0] != 'i1': continue # just a comment
			if noteLine[12] not in used['corpusId']:
				used['corpusId'][noteLine[12]] = []
			used['corpusId'][noteLine[12]].append(nidx)
			if noteLine[13] not in used['overlapId']:
				used['overlapId'][noteLine[13]] = []
			used['overlapId'][noteLine[13]].append(nidx)
			
		if ops.CSOUND_RENDER_SEPARATE_FILE_FOR_EACH_CORPUS_OVERLAP and len(used['overlapId']) > 1:
			for vcId, lines in used['overlapId'].iteritems():
				thisScoName = os.path.join(ops.TMP_PATH, 'overlap'+str(vcId)+'.sco')
				thisRenderOutput = path.locate(os.path.splitext(ops.CSOUND_OUTPUT_FILE)[0]+'-overlap'+str(vcId)+os.path.splitext(ops.CSOUND_OUTPUT_FILE)[1])
				tmpLines = []
				for lidx in lines: tmpLines.append(self.scoreNotes[lidx])
				outputListOfLists = utilities.printableList(tmpLines)
				tmpHandle = open(thisScoName, 'w') # open handle
				for strline in outputListOfLists:
					tmpHandle.write(strline)
				tmpHandle.close() # close handle
				ag.scoresToRender.append(csoundRenderObj(thisScoName, thisRenderOutput, len(outputListOfLists)))

		if ops.CSOUND_RENDER_SEPARATE_FILE_FOR_EACH_CORPUS_ENTRY and len(used['corpusId']) > 1:
			for vcId, lines in used['corpusId'].iteritems():
			
				thisScoName = os.path.join(ops.TMP_PATH, 'corpus'+str(vcId)+'.sco')
				thisRenderOutput = path.locate(os.path.splitext(ops.CSOUND_OUTPUT_FILE)[0]+'-corpus'+str(vcId)+os.path.splitext(ops.CSOUND_OUTPUT_FILE)[1])
				tmpLines = []
				for lidx in lines: tmpLines.append(self.scoreNotes[lidx])
				outputListOfLists = utilities.printableList(tmpLines)
				tmpHandle = open(thisScoName, 'w') # open handle
				for strline in outputListOfLists:
					tmpHandle.write(strline)
				tmpHandle.close() # close handle
				ag.scoresToRender.append(csoundRenderObj(thisScoName, thisRenderOutput, len(outputListOfLists)))
	############################################################################
	def writeParameterScore(self):
		fh = open(path.locate(ops.PARAM_SCORE_FILE), 'w')
		commentline = ';time file minSearchDist maxSearchDist param1 [param2 param3...]\n'
		fh.write(commentline)
		for (start, stop) in tgt.segFrames:
			ag.segAnal(tgt, descriptorData.getDescriptorObjByName('power-seg'), start, stop)
			scoreline = 'tgt-segment %f %f %f %f %f ; start, stop, peak, power, maxpower\n'%(start*ag.frameSec, stop*ag.frameSec, (start+tgt.data['raw']['peakTime-seg'][start])*ag.frameSec, tgt.data['raw']['power-seg'][start], tgt.maxPower)
			fh.write(scoreline)
		for t in range(tgt.timeInFrames):
			scoreline = 'tgt-amp %f %f ; time amp\n'%(t*ag.frameSec, tgt.data['orig']['power'][t]/tgt.maxPower)
			fh.write(scoreline)
		for hidx, (handle, thisTime, agAmp, semitones) in enumerate(self.noteHandles):
			if handle.hasParams:
				scoreline = 'segment %f %s %f %f %s'%(thisTime, handle.filename, self.minMaxSearch[hidx][0], self.minMaxSearch[hidx][1], ' '.join(handle.extraMappedData))
				fh.write(scoreline)
			else:
				scoreline = 'corpusSnd %s\n'%' '.join(self.handleToNote[handle][1:])
				fh.write(scoreline)
		fh.close()
	############################################################################
	def done(self, csoundBool, qlistBool, OmBool, midiFileBool):
		if csoundBool:
			self.filesToRender = self.writeScoreNotes()
		
		if qlistBool: self.writeQlistNotes()
		
		if ops.SAVE_OUTPUT_SEGMENT_LABELS:
			outputLabels = open(path.locate(ops.OUTPUT_SEGMENT_LABELS), 'w')
			self.outputLabelData.sort()
			for data in self.outputLabelData: outputLabels.write('\t'.join(data))
			outputLabels.close()
			
		self.scoreMixHandle.close()
		if ops.MAKE_MIDIFILE:
			import midifile
			MyMIDI = midifile.MIDIFile(1)
			MyMIDI.addTrackName(0, 0., "AudioGuide Track")
			MyMIDI.addTempo(0, 0., int(ops.TEMPO_CHANGE))
			tempoScaler = float(ops.TEMPO_CHANGE)/60.
			for thisTime, noteDuration, midiAmp, name, MIDIpitch, rmsPeak, peakTimeSec, effDur, onsetTime, offsetTime, handle, stretchCoeff in self.OmNotes:
				midiTime = thisTime*tempoScaler
				midiDuration = noteDuration*tempoScaler
				MyMIDI.addNote(0, 0, MIDIpitch, midiTime, midiDuration, midiAmp)
			
			# And write it to disk.
			binfile = open(self.midiFileName, 'wb')
			MyMIDI.writeFile(binfile)
			binfile.close()

		if ops.MAKE_PARAM_SCORE: self.writeParameterScore()


############################################################################
############################################################################
############################################################################
############################################################################
class csoundRenderObj:
	def __init__(self, scoName, outputName, numberEvents):
		self.scoName = scoName
		self.outputName = outputName
		self.events = numberEvents


def renderWithCsound(renderObjs, barString):
	import subprocess
	totalEvents = 0
	totalEventCounter = 0
	for rObj in renderObjs: totalEvents += rObj.events
	p.barOpen(barString, totalEvents-1)
	for rObj in renderObjs:
		csoundCommand = []
		if ops.CSOUND_RENDER_WITH_ULIMIT:
			csoundCommand.extend(['ulimit -n 2000 ;'])
		csoundCommand.extend(['csound','--output="'+rObj.outputName+'"', '"'+os.path.join(ops.AG_PATH,'scripts','csoundRender.orc')+'"', '"'+rObj.scoName+'"' ]	)
		cs = subprocess.Popen(' '.join(csoundCommand), shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
		output = open(os.path.join(ops.TMP_PATH, 'agCSOUNDLOG.txt'), 'w')
		while True:
			o = cs.stderr.readline()
			if o == '' and cs.poll() != None: break
			output.write(o)
			o = o.split()
			if len(o) < 3: continue
			if o[2] == 'giNoteCounter': 
				totalEventCounter += 1
				p.post('Note '+str(totalEventCounter)+'/'+str(totalEvents)+".", inc=True, forceInc=totalEventCounter)
	p.barClose("Done.")



class corpusEntry:
	def __init__(self, vcNumb, userData, globalAttributes):
		self.vc = vcNumb
		self.timeList = []
		self.attributeOrigins = {}
		### defaults
		self.segmentationFile = None
		self.MWtext = ''
		self.MWnotehead = '.'
		self.limit = {}
		self.concatFileName = None
		self.hasParams = False
		####
		if type(userData) == str: # then make a default dict
			self.name = userData
		elif type(userData) == dict: # add user-supplied dict elements
			self.name = userData['name']
			for item, val in userData[i].iteritems():
				setattr(self, item, val)
				self.attributeOrigins[item] = 'local'
		# add global attributes from CORPUS_GLOBAL_ATTRIBUTES
		for item, val in globalAttributes.iteritems():
			if item == 'limit':
				for i, v in val.iteritems():
					if not self.limit.has_key(i): self.limit[i] = v #local overrides global
			else:
				if not hasattr(self, item):
					setattr(self, item, val)
					self.attributeOrigins[item] = 'global'
		# fill in missing attributes from CORPUS_GLOBAL_ATTRIBUTES_DEFAULTS
		for item, val in ops.CORPUS_GLOBAL_ATTRIBUTE_DEFAULTS.iteritems():
			if item == 'limit':  continue
			if not hasattr(self, item):
				setattr(self, item, val)
				self.attributeOrigins[item] = 'default'




		print dir(self), self.attributeOrigins
		sys.exit()

################################################################################
################################################################################
################################################################################
class source:
	def __init__(self, sourceFromOpts):
		if len(ops.CORPUS) == 0: return # maybe just analyzing a target?
		printVerbose = False
		self.src = []
		self.corpusNames = []
		self.corpusPrintData = {}
		self.numberVc = 0
		self.simSelectRuleByCorpusId = {}
		self.corpusPocketingAssignment = {} # (start, stop): corpusID
		self.lastVoice = None
		# below for histograms
		self.fileSelectedList = []
		self.nameSelectedList = []
		self.voiceRestrictPerFrame = {}
		self.timesVoiceWasPlayed = {}
		self.vcToCorpusName = {}
		# FOR MUSICAL WRITING
		self.MWData = {} # same as ops.MW_DEFINE_INSTRUMENTS, but with default values added if not supplied
		for MWinstrName, dataDict in ops.MW_DEFINE_INSTRUMENTS:
			self.MWData[MWinstrName] = {'clef': 'G', 'midiChn': 1, 'key': "CM", 'polyphony': sys.maxint, 'shortestDur': False}
			# sys.maxint polyphony value because, by default, there is no limit to instrument overlap density!!!!
			for key, val in dataDict.iteritems():
				self.MWData[MWinstrName][key] = val

		srcCopy = ops.CORPUS[:] # making a copy to not overwrite original options
		for i in range(len(srcCopy)):
			#c = corpusEntry(i, srcCopy[i], ops.CORPUS_GLOBAL_ATTRIBUTES)
			self.numberVc += 1
			# make default attribute dict
			thisSourceEntry = {'vc': i, 'timeList': [], 'segmentationFile': None, 'MWtext': '', 'MWnotehead': '.', 'limit': {}, 'concatFileName': None, 'hasParams': False}
			thisSourceLimits = {} # descriptor-based limits
			globalOrLocal = {} # keep track of whether attributes are locally or globally assigned

			if type(srcCopy[i]) in [str, unicode]: # then make a default dict
				thisSourceEntry['name'] = str(srcCopy[i])
			elif type(srcCopy[i]) == dict: # add user-supplied dict elements
				thisSourceEntry['name'] = srcCopy[i]['name']
				for item, val in srcCopy[i].iteritems():
					thisSourceEntry[item] = val
					globalOrLocal[item] = 'local'
			if thisSourceEntry['concatFileName'] == None:
				 thisSourceEntry['concatFileName'] = thisSourceEntry['name']

			# add global attributes from CORPUS_GLOBAL_ATTRIBUTES
			for item, val in ops.CORPUS_GLOBAL_ATTRIBUTES.iteritems():
				if item == 'limit':
					for i, v in val.iteritems():
						if not thisSourceEntry['limit'].has_key(i): thisSourceEntry['limit'][i] = v #local overrides global
				else:
					if not thisSourceEntry.has_key(item): thisSourceEntry[item] = val #local overrides global
					globalOrLocal[item] = 'global'
			# fill in missing attributes from CORPUS_GLOBAL_ATTRIBUTES_DEFAULTS
			for item, val in ops.CORPUS_GLOBAL_ATTRIBUTE_DEFAULTS.iteritems():
				if item == 'limit': thisSourceEntry['limit'].update(val)
				if not thisSourceEntry.has_key(item): thisSourceEntry[item] = val
			
			# set corpusID restrict by selection rule for later use in evaluateValidHandles...
			self.simSelectRuleByCorpusId[i] = thisSourceEntry['simSelectRule']
	
			#if thisSourceEntry['name'] in self.corpusNames:
			#	utilities.error("ops", "What were you thinking?  You cannot use the same corpus name more than once -> "+thisSourceEntry['name'])
			self.corpusNames.append(thisSourceEntry['name'])

			exists, srcpath, fileType = path.test(thisSourceEntry['name'], ops.SOURCE_PATH) # user path, relative path
			# use non-standard location segmentation file?
			if thisSourceEntry['segmentationFile'] == None: # add default name of segmentation file if nothing is specified by the user			
				thisSourceEntry['segmentationFile'] = [path.locate(srcpath+thisSourceEntry['segmentationExtension'])]
			elif type(thisSourceEntry['segmentationFile']) == str: # if a single string
				thisSourceEntry['segmentationFile'] = [path.test(thisSourceEntry['segmentationFile'], ops.SOURCE_PATH)[1]]
			elif type(thisSourceEntry['segmentationFile']) in [tuple, list]: # a list of seg files
				tmp = []
				for string in thisSourceEntry['segmentationFile']:
					tmp.append(path.test(string, ops.SOURCE_PATH)[1])
				thisSourceEntry['segmentationFile'] = tmp
			if not exists:
				utilities.error("ops", "Bad Source input name -> "+thisSourceEntry['name']+'.  Not a file, a directory')
			
			if fileType == 'file': # an audio file
				##################
				## input a FILE ## -- look for audacity-style txt label file
				##################
				times = []
				for segFile in thisSourceEntry['segmentationFile']:
					if not thisSourceEntry['wholeFile'] and not os.path.isfile(segFile):
						utilities.error('segmentation file', "Cannot find segmentation file '"+segFile+"'")
					segs = utilities.readAudacityLabelFile(segFile, thisSourceEntry['wholeFile'], printVerbose)
					times.extend(utilities.readAudacityLabelFile(segFile, thisSourceEntry['wholeFile'], printVerbose))
				for timeSeg in times:
					writeLine = [srcpath]
					writeLine.extend(timeSeg)
					thisSourceEntry['timeList'].append( writeLine )
				thisSourceEntry['mapPitch'] = {} # REMOVE
			elif fileType == 'dir': # a directory
				#######################
				## input a DIRECTORY ##
				#######################
				files = utilities.getDirListOnlyExt(srcpath, thisSourceEntry['recursive'], ops.SOUNDFILE_EXTENSIONS)
				thisSourceEntry['segmentationFile'] = None # don't print it
				for file in files:
					segFileTest = path.locate(file)+thisSourceEntry['segmentationExtension']
					if not thisSourceEntry['wholeFile'] and not os.path.isfile(segFileTest):
						utilities.error('segmentation file', "Cannot find segmentation file '"+segFileTest+"'.  To specify the use of whole sound files as corpus segments, write this corpus entry as: \n\t{'name': '"+thisSourceEntry['name']+"', 'wholeFile': True}\nor\n\tCORPUS_GLOBAL_ATTRIBUTES = {'wholeFile': True}")
					times = utilities.readAudacityLabelFile(path.locate(file)+thisSourceEntry['segmentationExtension'], thisSourceEntry['wholeFile'], printVerbose)
					for timeSeg in times:
						writeLine = [file]
						writeLine.extend(timeSeg)
						thisSourceEntry['timeList'].append( writeLine )
					thisSourceEntry['mapPitch'] = []
				thisSourceEntry['numbSfFiles'] = len(files)
			# reset counters...
			segCount = 0
			windowDist = metrics.hannWin(len(thisSourceEntry['timeList'])*2)
			# add limits per corpus entry
			if not thisSourceEntry.has_key('limit'):
				thisSourceEntry['limit'] = {}
			else:
				for key, val in thisSourceEntry['limit'].iteritems():
					if type(val) != dict:
						val = {'val': val}
						thisSourceEntry['limit'][key] = val
					thisSourceEntry['limit'][key]['scope'] = "local" # add scoping
			# add "global" limits to each corpus entry from LIMIT_CORPUS_BY_DESCRIPTOR
			for key, val in ops.LIMIT_CORPUS_BY_DESCRIPTOR.iteritems():
				# only if the corpus entry doesn't already have this item
				# otherwise local limits take precedence!
				if type(val) == tuple:
					val = {'low': val[0], 'high': val[1]}
				if not thisSourceEntry['limit'].has_key(key):
					thisSourceEntry['limit'][key] = val
					thisSourceEntry['limit'][key]['scope'] = "global" # add scoping
			# restrictInTime if a non-zero positive float/int
			if thisSourceEntry['restrictInTime'] > 0:
				self.voiceRestrictPerFrame[i] = int(thisSourceEntry['restrictInTime']/ag.frameSec)
			# segment list
			for idx in range(len(thisSourceEntry['timeList'])): 
				file = thisSourceEntry['timeList'][idx][0]
				filehead = os.path.split(file)[1]
				start = thisSourceEntry['timeList'][idx][1]
				end = thisSourceEntry['timeList'][idx][2]
				# matchSting: includeString/excludeString
				if thisSourceEntry.has_key('includeString') and thisSourceEntry['includeString'] not in ['', None]:
					skip = True
					if type(thisSourceEntry['includeString']) not in [list, tuple]: thisSourceEntry['includeString'] = [thisSourceEntry['includeString']]
					for test in thisSourceEntry['includeString']:
						if utilities.matchString(filehead, test, caseSensative=True): skip = False
					if skip: continue
				if thisSourceEntry.has_key('excludeString') and thisSourceEntry['excludeString'] not in ['', None]:
					skip = False
					if type(thisSourceEntry['excludeString']) not in [list, tuple]: thisSourceEntry['excludeString'] = [thisSourceEntry['excludeString']]
					for test in thisSourceEntry['excludeString']:
						if utilities.matchString(filehead, test, caseSensative=True): skip = True
					if skip: continue
				#        minTime / maxTime
				if start < thisSourceEntry['minTime']: continue # skip it
				if start > thisSourceEntry['maxTime']: continue # skip it
				# matchTime: includeTimes/excludeTimes
				if thisSourceEntry.has_key('includeTimes') and thisSourceEntry['includeTimes'] not in ['', None]:
					skip = True
					if type(thisSourceEntry['includeTimes']) not in [list, tuple]: thisSourceEntry['includeTimes'] = [thisSourceEntry['includeTimes']] # force it to be a list
					for timeTuple in thisSourceEntry['includeTimes']:
						if start >= timeTuple[0] and end <= timeTuple[1]: skip = False
					if skip: continue
				if thisSourceEntry.has_key('excludeTimes') and thisSourceEntry['excludeTimes'] not in ['', None]:
					skip = False
					if type(thisSourceEntry['excludeTimes']) not in [list, tuple]: thisSourceEntry['excludeTimes'] = [thisSourceEntry['excludeTimes']] # force it to be a list
					for timeTuple in thisSourceEntry['excludeTimes']:
						if start >= timeTuple[0] and end <= timeTuple[1]: skip = True
					#print filehead, start, end, skip
					if skip: continue
					

					#sys.exit()
				distance = 1
				metaData = thisSourceEntry['timeList'][idx][3:] # the rest
				# for include/exclude based on label file meta data
				if thisSourceEntry.has_key('includeMetaData'):
					loadMe = True
					if metaData[0] not in thisSourceEntry['includeMetaData']: continue
				if thisSourceEntry.has_key('excludeMetaData'):
					loadMe = True
					for meta in metaData:
						if meta in thisSourceEntry['excludeMetaData']: loadMe = False
					if not loadMe: continue
				if len(thisSourceEntry['timeList'][idx]) >= 3:
					extraMappedData = thisSourceEntry['timeList'][idx][3:]
				else:
					extraMappedData = ''
				# weight likelihood of selection
				if thisSourceEntry.has_key('weight'):
					if type(thisSourceEntry['weight']) in [int, float]:
						distance = distance*thisSourceEntry['weight']
				# set this sample's transposition method
				if thisSourceEntry.has_key('transMethod'):
					if thisSourceEntry['transMethod'][0] == 'markov' and thisSourceEntry['transMethod'][1] == 'filename':
						transMethod = [thisSourceEntry['transMethod'][0], metrics.MIDIPitchByFileName(file, None)] # infer pitch from filename
				else: 
					if ops.RANDOM_TRANSPOSE == None: transMethod = ["none"]
					if ops.SMART_TRANSPOSE: transMethod = ["smart"]
					elif ops.TRANSPOSE_TO_FIT_TARGET_SEGMENT_LENGTH: transMethod = ["matchTgtTemporality"]
					elif type(ops.RANDOM_TRANSPOSE) is list: transMethod = ["randomRange", ops.RANDOM_TRANSPOSE]
				# test if limitDur is set...
				if thisSourceEntry.has_key('limitDur'):
					if end == None: end = thisSourceEntry['limitDur']
					elif end-start > thisSourceEntry['limitDur']: end = start+thisSourceEntry['limitDur']
				# dynamic from filename
				#print file, utilities.getDynamicFromFilename(file, notFound=None)
				
				self.src.append({'file': file, 'startSec': start, "endSec": end, 'scaleDistance': distance, 'voiceID': thisSourceEntry['vc'], 'arrayName': thisSourceEntry['name'], 'transMethod': transMethod, 'segNumber': segCount, 'scaleDb': thisSourceEntry['scaleDb'], 'mappedPitch':-1, 'extraMappedData':extraMappedData, 'limit': thisSourceEntry['limit'], 'MWinstrName': thisSourceEntry['MWinstrName'], 'MWtext': thisSourceEntry['MWtext'], 'MWnotehead': thisSourceEntry['MWnotehead'], 'restrictOverlaps': thisSourceEntry['restrictOverlaps'], 'concatFileName': thisSourceEntry['concatFileName'], 'onsetLen': thisSourceEntry['onsetLen'], 'offsetLen': thisSourceEntry['offsetLen'], 'midiPitchMethod': thisSourceEntry['MIDIPitch'], 'hasParams': thisSourceEntry['hasParams']})
				segCount += 1
				###########################
				## done with CORPUS loop ##
				###########################
			if not thisSourceEntry.has_key('numbSfFiles'): thisSourceEntry['numbSfFiles'] = 1
			self.corpusPrintData[thisSourceEntry['name']] = {'segs': str(segCount), 'fileType': fileType, 'limit': thisSourceEntry['limit'], 'limitStart': thisSourceEntry['minTime'], 'limitEnd': thisSourceEntry['maxTime'], 'numbSfFiles': thisSourceEntry['numbSfFiles'], 'restrictInTime': thisSourceEntry['restrictInTime'], 'segFile': thisSourceEntry['segmentationFile'], 'restrictOverlaps': thisSourceEntry['restrictOverlaps'], 'scaleDb': thisSourceEntry['scaleDb']}
			self.vcToCorpusName[thisSourceEntry['vc']] = thisSourceEntry['name']
		self.pool = {}
		self.preLimitTotal = len(self.src)
		for i in range(self.numberVc):
			self.pool[i] = {}
	############################################################################
	def nameTest(self, name, prefixList): # return only same with a certain starting prefix
		test = True
		name = name.split('/')[-1]
		for t in prefixList:
			if name[0:len(t)] == t:
				test = True
				break
			else:
				test = False
		return test
	############################################################################
	def initConcatenate(self, tgtLength, allSourceHandles):
		for handle in ag.sourceHandles:
			self.pool[handle.voiceID][handle] = 0 # clear time-dependant pool
		# for target seg increment...
		counter = 0
		self.possibleVoicesPerFrame = []
		for i in range(0, tgt.lengthInFrames): self.possibleVoicesPerFrame.append([])
		for pair in tgt.segFrames:
			for i in range(pair[0], pair[1]):
				self.possibleVoicesPerFrame[i].append(counter%src.numberVc)
			counter += 1
		# to count overlaps of each corpus voice
		self.corpusVcOverlapCounter = {}
		for vc in range(self.numberVc): self.corpusVcOverlapCounter[vc] = np.zeros(tgt.lengthInFrames, dtype=int)
	############################################################################
	def updateSourcePool(self, chosenHandle):
		if ops.CAN_USE_SAME_SEGMENT_TWICE and ops.CAN_REUSE_SEGMENT_AFTER > 0:
			self.pool[chosenHandle.voiceID][chosenHandle] = int(tgt.timeInFrames+((ops.CAN_REUSE_SEGMENT_AFTER/ag.frameSec)))
		elif ops.CAN_USE_SAME_SEGMENT_TWICE == False: 
			del self.pool[chosenHandle.voiceID][chosenHandle] # remove it altogether


	############################################################################
	def updateTargetPocketCorpusRestriction(self, tgtTime, tgtSegs, chosenHandle):
		# first check to see if this selection already has a pcoket craved out for it

		for st, ed in self.corpusPocketingAssignment:
			if tgtTime >= st and tgtTime <= ed:
				return
		
		# it does not, so make one!
		for sidx, (st, ed) in enumerate(tgtSegs):
			if tgtTime >= st and tgtTime <= ed:
				break
		pocketStartFrame = tgtSegs[sidx][0]
		numberOfSegmentsToPocket = random.choice(range(int(ops.CORPUS_POCKET_SEGMENT_RANDOM_RANGE[0]), int(ops.CORPUS_POCKET_SEGMENT_RANDOM_RANGE[1])))
		#numberOfSecondsToPocket = random.uniform(ops.CORPUS_POCKET_SEGMENT_RANDOM_RANGE[0], ops.CORPUS_POCKET_SEGMENT_RANDOM_RANGE[1])
		#pocketEndFrame = int(pocketStartFrame+(numberOfSecondsToPocket/ag.frameSec))
		pocketEndFrame = tgtSegs[sidx:sidx+numberOfSegmentsToPocket][-1][1]
		self.corpusPocketingAssignment[(pocketStartFrame, pocketEndFrame)] = chosenHandle.voiceID
		print "VOICE %i solo pocket for %i segments"%(chosenHandle.voiceID, numberOfSegmentsToPocket)
		print '\n\n'


	############################################################################
	def evaluateValidSamples(self, maxActiveVoices):
		validHandles = []
		validSteams = []
		# VOICE_WEIGHT - a dict where name is key and val is percentage likelihood.
		#    percentages don't have to add up to 1 or 100, but a summed
		voiceWeightWinner = None
		if ops.VOICE_WEIGHT not in [{}, None]:
			totalWeights = 0.
			for k, v in ops.VOICE_WEIGHT.iteritems(): totalWeights += v
			tmpNormWeights = []
			lastVal = 0
			for k, v in ops.VOICE_WEIGHT.iteritems():
				tmpNormWeights.append(((lastVal, (v/totalWeights)+lastVal), k))
				lastVal = (v/totalWeights)+lastVal
			randVal = random.random()
			for tupleVals, key in tmpNormWeights:
				if randVal >= tupleVals[0] and randVal < tupleVals[1]:
					voiceWeightWinner = key
					break

		for vc, pool in self.pool.iteritems():
			# methods for restricting voices based upon the last selected voice
			if src.lastVoice != None:
				# ROTATE_VOICES is a bool
				if ops.ROTATE_VOICES: 
					if vc != (src.lastVoice+1)%src.numberVc: continue # skip this set of samples
				# VOICE_PATTERN
				if ops.VOICE_PATTERN not in [None, []]: # not case sensative!, oes a search, so matches partial strings
					if not utilities.matchString(src.vcToCorpusName[vc], ops.VOICE_PATTERN[len(ag.selectedSegments)%len(ops.VOICE_PATTERN)], caseSensative=False): continue
				# VOICE_WEIGHT, made above
				if voiceWeightWinner != None:
					if not utilities.matchString(src.vcToCorpusName[vc], voiceWeightWinner, caseSensative=False): continue
			# do voice restriction per second
			passVcRestriction = True
			if src.timesVoiceWasPlayed.has_key(vc) and src.voiceRestrictPerFrame.has_key(vc):
				for playTime in src.timesVoiceWasPlayed[vc]:
					if abs(tgt.timeInFrames-(playTime/2)) <= src.voiceRestrictPerFrame[vc]:
						passVcRestriction = False
						break
			if not passVcRestriction: continue
			# do voice restriction per segment using self.possibleVoicesPerFrame
			if ops.VOICE_RESTRICT_PER_SEGMENT and vc not in self.possibleVoicesPerFrame[tgt.timeInFrames] and len(self.possibleVoicesPerFrame[tgt.timeInFrames]) > 0: continue

			# voice pocketing -- permits only one corpus voice per random grouping of target segments
			#	ops.CORPUS_POCKETING = True
			#	ops.CORPUS_POCKET_SEGMENT_RANDOM_RANGE = [1, 4] # in seconds
			#	ops.CORPUS_POCKET_ONLY_SELECTION_NUMBER = [1, 2, 3] # which sim selections
			#	uses self.corpusPocketingAssignment as a registry to keep track
			#		self.corpusPocketingAssignment = {} # (start, stop): corpusID
			if ops.CORPUS_POCKETING:
				CORPUS_POCKET_WHICH_SIM_SELECTS = [1]
				if tgt.segmentsPerFrame[tgt.timeInFrames] not in CORPUS_POCKET_WHICH_SIM_SELECTS:
					
					prev_pocket_vc = None
					closest_prev_pocket = []
					has_pocket = False
					for (start, stop), vcId in self.corpusPocketingAssignment.iteritems():
						if start <= tgt.timeInFrames and stop >= tgt.timeInFrames:
							has_pocket = True
							
						if stop < tgt.timeInFrames: closest_prev_pocket.append([tgt.timeInFrames-stop, vcId])
					closest_prev_pocket.sort()
					if has_pocket:
						if vcId != vc: 
							continue
					else:
						if ops.CORPUS_POCKETING_ROTATE and len(closest_prev_pocket) > 0:
							if len(closest_prev_pocket) == 1:
								print "ONLY ONE LAST POCKET", closest_prev_pocket[0][1]
								invalidForRotation = [closest_prev_pocket[0][1]]
							else:
								print "TWO LAST POCKETS", closest_prev_pocket[0][1], closest_prev_pocket[1][1], 
								invalidForRotation = [closest_prev_pocket[0][1], closest_prev_pocket[1][1]]
							if vc in invalidForRotation: continue
	
	
				#print "\n\n"
			# restrict corpus ID by number of selected overlapping samples
			# uses eval() to test agains overlap number...
			if self.simSelectRuleByCorpusId.has_key(vc) and self.simSelectRuleByCorpusId[vc] != None:
				if not eval(str(tgt.segmentsPerFrame[tgt.timeInFrames])+self.simSelectRuleByCorpusId[vc]): continue
			###################################################
			## loop through each sound segment in this voice ##
			###################################################
			for sampleHandle, time in pool.iteritems(): # add valid samples to the sample list for this frame
				if sampleHandle.restrictOverlaps != None: # limiting each corpus entry by the number of simultanous overlaps..
					voicesPerFrame = self.corpusVcOverlapCounter[vc][tgt.timeInFrames:tgt.timeInFrames+sampleHandle.data['raw']['effDur-seg'][0]]
					#bach noteprint "\n\nVOICES", np.max(voicesPerFrame), "\n\n"
					if np.max(voicesPerFrame) >= sampleHandle.restrictOverlaps: continue
					#print "PASSED", sampleHandle.name, sampleHandle.restrictOverlaps, np.max(voicesPerFrame)
				# if the sample is valid to be picked...
				if time <= tgt.timeInFrames: 
					availableStreams = src.timeRestrictTest(sampleHandle) # tests based upon time restrictions
					if len(availableStreams) > 0:
						validHandles.append(sampleHandle) # add it if you've made it this far
						validSteams.append(availableStreams)
		return validHandles, validSteams
	############################################################################
	def initTimeVaryingRestrict(self):
		if len(ops.CORPUS_RESTRICT_DISCRIPTOR_IN_TIME) > 0: self.restrictTimeVarying = True
		else: self.restrictTimeVarying = False
		if self.restrictTimeVarying:
			self.sourceTimeRestrictAnal = {}
			for restrict in ops.CORPUS_RESTRICT_DISCRIPTOR_IN_TIME:
				minVal, maxVal = getNormCoefs(ag.sourceHandles, restrict['anal'], 'minmax', None, None)
				self.sourceTimeRestrictAnal[restrict['anal']] = {}
				if restrict.has_key('inits') and len(restrict['inits']) >= ops.CORPUS_RESTRICT_NUMBER_STREAMS:
					initVals = restrict['inits']
				else: 
					initVals = [None]*ops.CORPUS_RESTRICT_NUMBER_STREAMS # no init if none supplied
				for i in range(ops.CORPUS_RESTRICT_NUMBER_STREAMS):
					self.sourceTimeRestrictAnal[restrict['anal']][i] = {'lastActive': 0, 'lastMetircVal': initVals[i], 'minVal': minVal, 'maxVal': maxVal, 'restrictionReleaseSec': restrict['restrictionReleaseSec'], 'percentRestriction': restrict['percentRestriction']}
			# data histogram
				tmpFeatureList = []
				for h in ag.sourceHandles:
					tmpFeatureList.append(h.data['raw'][restrict['anal']][0])
				tmpSlicedList = sliceListIntoEvenlySpacedLists(tmpFeatureList, (maxVal-minVal)/200.0, minVal, maxVal+10)
				distribution = []
				for i in tmpSlicedList: distribution.append(len(i))
			# printing
			for i, v in self.sourceTimeRestrictAnal.iteritems():
				print i, '->', v
	############################################################################
	def updateRestrict(self, chosenSampleHandle, chosenSampleData):
		# voice per second restriction...
		if not self.timesVoiceWasPlayed.has_key(chosenSampleHandle.voiceID):
			self.timesVoiceWasPlayed[chosenSampleHandle.voiceID] = []
		self.timesVoiceWasPlayed[chosenSampleHandle.voiceID].append(tgt.timeInFrames)
		# time varying restriction...
		if len(ops.CORPUS_RESTRICT_DISCRIPTOR_IN_TIME) > 0:
			pickedStream = chosenSampleData['availableStreams'][0]
			for anal in self.sourceTimeRestrictAnal: # record for each user-supplied feature...
				self.sourceTimeRestrictAnal[anal][pickedStream]['lastMetircVal'] = chosenSampleHandle.data['raw'][anal][0]
				self.sourceTimeRestrictAnal[anal][pickedStream]['lastActive'] = tgt.timeInFrames
	############################################################################
	def timeRestrictTest(self, sampleHandle):
		if not self.restrictTimeVarying:
			return [0] # returning stream 0, so everything is enabled
		# if passing this point, perform restriction test...
		availableStreams = []
		for anal in self.sourceTimeRestrictAnal: # test for each user-supplied feature...
			for streamID in self.sourceTimeRestrictAnal[anal]: # test for each stream...
				if self.sourceTimeRestrictAnal[anal][streamID]['lastMetircVal'] == None: 
					availableStreams.append(streamID) # don't restrict if not set yet by updateRestrict()
				else: # then test restriction...
					timeDifference = (tgt.timeInFrames-self.sourceTimeRestrictAnal[anal][streamID]['lastActive'])*ag.frameSec
					timeRatioToOnsets = timeDifference/self.sourceTimeRestrictAnal[anal][streamID]['restrictionReleaseSec']
					percentPossibleDueToTime = interpolateVal(timeRatioToOnsets, self.sourceTimeRestrictAnal[anal][streamID]['percentRestriction'], 100)
					if percentPossibleDueToTime >= 100: # then its over 100%, definitely do it
						availableStreams.append(streamID) 
						continue
					minVal = float(self.sourceTimeRestrictAnal[anal][streamID]['minVal'])
					maxVal = float(self.sourceTimeRestrictAnal[anal][streamID]['maxVal'])
					lastInThisStream = self.sourceTimeRestrictAnal[anal][streamID]['lastMetircVal']
					thisSample = sampleHandle.data['raw'][anal][0]
					absRawDiff = abs(lastInThisStream-thisSample) # difference of two metrics as a raw value
					absRawDiffAsPercentOfRange = (absRawDiff/(maxVal-minVal))*100.0
					if absRawDiffAsPercentOfRange <= percentPossibleDueToTime: # then its within the window, so add it
					#	print "PASSED", sampleHandle.data['raw'][anal][0], absRawDiffAsPercentOfRange, percentPossibleDueToTime
						availableStreams.append(streamID)
					#else: 
					#	print "FFFFFF", sampleHandle.data['raw'][anal][0], absRawDiffAsPercentOfRange, percentPossibleDueToTime
		#	print len(self.sourceTimeRestrictAnal)
		return availableStreams
	############################################################################
	def printCorpusInfo(self):
		postLevel = 2
		printNumbSegsOrFiles = True
		printLimits = True
		# get number of total potential segments
		totalSegs = 0
		for name, dict in self.corpusPrintData.iteritems(): totalSegs += int(dict['segs'])
		p.middle('Corpus - '+str(totalSegs)+' Possible Segments', postLevel)
		
		for name, dict in self.corpusPrintData.iteritems():
			if dict['segFile'] != None: # make a printable list of segmentation files
				tmp = []
				for item in dict['segFile']:
					tmp.append(os.path.split(item)[1])
				name = ','.join(tmp)
			
			end = dict['limitEnd']
			if end == sys.maxint: end = 'end'
			printStr = "${YELLOW}"+str(dict['fileType'])+"${NORMAL} "+str(name)+" -> "
			if printNumbSegsOrFiles:
				if dict['fileType'] == 'file':
					printStr += "${RED}"+dict['segs']+"${NORMAL} segments"
					if dict['limitStart'] != 0 and str(end) not in ["end", None, "none"]:
						printStr += " from "+str(dict['limitStart'])+" to "+str(end)+"."
				elif dict['fileType'] == 'dir': printStr += "${RED}"+dict['segs']+"${NORMAL} segments from "+str(dict['numbSfFiles'])+" files."
				elif dict['fileType'] == 'name': printStr += "${RED}"+dict['segs']+"${NORMAL} segments"
				
			if printLimits:
				if len(dict['limit']) > 0:
					for key, valDict in dict['limit'].iteritems():
						restrict = [None, None]
						for item, value in valDict.iteritems():
							if item == 'scope': scoped = value
							if item == 'low': restrict[0] = str(value)
							if item == 'high': restrict[1] = str(value)
							if item == 'lowPer': restrict[0] = str(value)+"%"
							if item == 'highPer': restrict[1] = str(value)+"%"
						if restrict[1] == None: restStr = "above "+str(restrict[0])
						elif restrict[0] == None: restStr = "below "+restrict[1]
						else: restStr = "between "+restrict[0]+" -> "+restrict[1]
						printStr += "\n\t${RED}"+key+ "${NORMAL} "+restStr+", scoped ${YELLOW}"+scoped+"ly${NORMAL}"
				if dict.has_key('restrictInTime') and dict['restrictInTime'] > 0:
					print "\tRestricting segments to once every", dict['restrictInTime'], "seconds."
				if dict['restrictOverlaps'] != None:
					printStr += "\n\tRestricting segments to a ${RED}maximum simultanous overlap${NORMAL} of "+str(dict['restrictOverlaps'])+" -- scoped ${YELLOW}"+dict['restrictOverlapsReason']+"ly${NORMAL}."

			p.post(printStr, postLevel)
		p.post( "" )
		pass
	############################################################################
	def printSampleHistogram(self):
		postLevel = 2
		noteHist = utilities.histogram(self.nameSelectedList)
		if len(noteHist) <= 1: noteHist = utilities.histogram(self.fileSelectedList)
		if ag.countRanOutOfSamples > 0: noteHist.append((ag.countRanOutOfSamples, "Out of Valid Segments"))
		p.middle('File Selection Histogram', postLevel)
		for (val, key) in noteHist:
			p.post( key+" -> ${RED}"+str(utilities.trunc((val/float(outputFiles.totalOutputNotes+ag.countRanOutOfSamples))*100, 2))+"%${NORMAL}", postLevel)
		p.post("", postLevel)
	############################################################################
	def printSimultanousHistogram(self):
		postLevel = 2
		tmpy = []
		for numbSeg in tgt.onsetsPerFrame:
			if numbSeg > 0: tmpy.append(numbSeg)
		#print tmpy
		noteHist = utilities.histogram(tmpy)
		#print noteHist
		p.middle('Simultaneous Selection Histogram', postLevel)
		for (val, key) in noteHist:
			p.post(str(key)+" note(s) -> ${RED}"+str(utilities.trunc(val/float(len(tmpy))*100, 2))+"%${NORMAL}", postLevel)
		p.post("", postLevel)
