import numpy as np
np.seterr(invalid='ignore')
import sys
import utilities
import simcalc


try:
	from aubio.aubioclass import *
except ImportError, err:
	pass

def set_all_segmented_methods(list):
	global ALL_SEG_METHDS
	ALL_SEG_METHDS = list

def flux(FFTs, maxs, FLUX_LAST_FFT_FRAME):
	# un-normalised flux
	if FLUX_LAST_FFT_FRAME == None: FLUX_LAST_FFT_FRAME = np.zeros(FFTs.shape[1])
	output = np.empty(FFTs.shape[0])
	output[0] = np.sum(np.power((FFTs[0]-FLUX_LAST_FFT_FRAME), 2)) # flux of last frame
	for i in range(1, FFTs.shape[0]):
		output[i] = np.sum(np.power((FFTs[i]-FFTs[i-1]), 2)) # flux between frames
	return output

def smooth(x, window_len=10, window='hanning'):
	"""smooth the data using a window with requested size.
	This method is based on the convolution of a scaled window with the signal.
	The signal is prepared by introducing reflected copies of the signal
	(with the window size) in both ends so that transient parts are minimized
	in the begining and end part of the output signal.
	input:
        x: the input signal 
        window_len: the dimension of the smoothing window
        window: the type of window from 'flat', 'hanning', 'hamming', 'bartlett', 'blackman'
            flat window will produce a moving average smoothing.

    output:
        the smoothed signal
        
    example:

    import numpy as np    
    t = np.linspace(-2,2,0.1)
    x = np.sin(t)+np.random.randn(len(t))*0.1
    y = smooth(x)
    
    see also: 
    
    numpy.hanning, numpy.hamming, numpy.bartlett, numpy.blackman, numpy.convolve
    scipy.signal.lfilter
 
    TODO: the window parameter could be the window itself if an array instead of a string   
    """
	if x.ndim != 1:
		raise ValueError, "smooth only accepts 1 dimension arrays."
	if x.size < window_len:
		raise ValueError, "Input vector needs to be bigger than window size."
	if window_len < 3:
		return x
	if not window in ['flat', 'hanning', 'hamming', 'bartlett', 'blackman']:
		raise ValueError, "Window is on of 'flat', 'hanning', 'hamming', 'bartlett', 'blackman'"
	s=np.r_[2*x[0]-x[window_len:1:-1], x, 2*x[-1]-x[-1:-window_len:-1]]
    #print(len(s))
	if window == 'flat': #moving average
		w = np.ones(window_len,'d')
	else:
		w = getattr(np, window)(window_len)
	y = np.convolve(w/w.sum(), s, mode='same')
	return y[window_len-1:-window_len+1]




##############################
## SINGLE-NUMBER MORPHOLOGY ##
##############################
def peakTimeSeg(powers):
	return [np.argmax(powers)+0.5] # the middle of the peak window, reduces error to +/-0.5


def descriptorSlope(dataHandle, descriptName, start, NORMALIZE_FOR_DURATION=True):
	# with Norbert Schnell
	if start == None: effStart = 0
	else: effStart = start
	effDur = dataHandle['effDur-seg'][effStart]
	descript = dataHandle[descriptName][effStart:effStart+effDur]
	start = int((len(descript))*0.2) # only take descriptor from %20-%80 
	end = int((len(descript))*0.8)
	values = np.array(descript[start:end])
	if len(values) < 2: return [0.0] # nada
	if NORMALIZE_FOR_DURATION:
		cmpLine = np.linspace(0, 1, num=len(values))
	else:
		cmpLine = np.linspace(0, len(values), num=len(values))
	return [np.polyfit(cmpLine, values, 1)[0]]


def MIDIPitchByFileName(name, midiPitchMethod, handle):
	# midiPitchMethod == 'filename' || int/float || yin-seg || centroid-seg
	if midiPitchMethod == 'filename':
		import os
		fileHead = os.path.splitext(os.path.split(name)[1])[0]
		return getMidiPitchFromString(fileHead)
	elif type(midiPitchMethod) in [float, int]:
		return midiPitchMethod
	elif midiPitchMethod == 'centroid-seg':
		return utilities.frq2Midi(handle.data['raw']['centroid-seg'][0])
	elif midiPitchMethod == 'f0-seg':
		return handle.data['raw']['f0-seg'][0]


FILENAME_PITCH_DICT = {} # save string results here so it doesn't have to reevaulate..
def getMidiPitchFromString(string):
	import os, re
	global FILENAME_PITCH_DICT
	if FILENAME_PITCH_DICT.has_key(string): return FILENAME_PITCH_DICT[string]
	pitchDict = {"c": 0,"cs": 1,"db": 1,"d": 2,"ds": 3,"eb": 3,"e": 4, 'fb': 4, "f": 5,"fs": 6,"gb": 6,"g": 7,"gs": 8,"ab": 8,"a": 9,"as": 10,"bb": 10, "b": 11, 'cb': 11}
	results = []
	results.extend(re.compile(r'[a-gA-G][sb\#][0123456789]').findall(string)) # re command for searching three element pattern
	results.extend(re.compile(r'[a-gA-G][0123456789]').findall(string)) # no accidentals!!
	if len(results) > 0: # found something
		result = results[0]
		if len(result) == 2: 
			pch = result[0].lower()
			octv = result[1]
		else:
			pch = result[0:2].lower().replace('#', 's')
			octv = result[2]
		returnPitch = pitchDict[pch]+(int(octv)+1)*12
		FILENAME_PITCH_DICT[string] = returnPitch
		return returnPitch
	else:
		FILENAME_PITCH_DICT[string] = -1
		return -1
	


def percentageInFile(handle, start, end):
	if handle.typed == 'target':
		output = [(start/float(handle.lengthInFrames))*100] # computed in frames
	else:
		output = [(handle.startSec/float(readsf.fileDuration(handle.filename)))*100] # computed in seconds
	return output



def effectiveDur(handle, time, ag):
	# RETURN SEGMENT LENGTH IN FRAMES
	if time == None: time = 0
	length = len(handle.data['raw']['power'])
	if handle.typed == 'target':
		start = time+ag.minSegLength
		if start >= length:  # if shorter than user specified segment time...
			handle.lastSegmentTerminated = 'fileEnd'
			return length-time
		else:
			stop = start+1
			if ag.minSegLength > 1: # get max in this time
				maxPowerInMinTime = max(handle.data[ops.TARGET_OFFSET_ENVELOPE]['power'][time:time+ag.minSegLength])
			else: # just use first value
				maxPowerInMinTime = handle.data[ops.TARGET_OFFSET_ENVELOPE]['power'][time]
			# drop!!!! important
			if ops.TARGET_SEGMENT_LENGTH_DROP_METHOD == 'relative':
				stopVal = maxPowerInMinTime/ops.TARGET_SEGMENT_DROP_RATIO #drop value determined from the peak amp in time->time+self.minSegLength
			elif ops.TARGET_SEGMENT_LENGTH_DROP_METHOD == 'absDb':
				stopVal = handle.absDbDropPower
			while stop+1 < length:
				if handle.data[ops.TARGET_OFFSET_ENVELOPE]['power'][stop] <= stopVal:
					stop += 1 # use next frame too, since looking ahead...
					handle.lastSegmentTerminated = ['drop', utilities.trunc(handle.data[ops.TARGET_OFFSET_ENVELOPE]['power'][stop], 2), str(stopVal)]
					break
				elif handle.data[ops.TARGET_OFFSET_ENVELOPE]['power'][stop+1]/handle.data[ops.TARGET_OFFSET_ENVELOPE]['power'][stop] >= ops.TARGET_OFFSET_RISE_RATIO:
					handle.lastSegmentTerminated = ['rise', utilities.trunc(handle.data[ops.TARGET_OFFSET_ENVELOPE]['power'][stop+1]/handle.data[ops.TARGET_OFFSET_ENVELOPE]['power'][stop], 2), str(ops.TARGET_OFFSET_RISE_RATIO)]
					break
				else:
					stop += 1 # keep looking
			if stop-time > ag.maxSegLength: # if longer than user specified maximum segment time...
				handle.lastSegmentTerminated = 'maxSegLength'
				return ag.maxSegLength
			else:
				# return the stopping point
				return stop-time
	################
	## for corpus ##
	################
	elif handle.typed == 'source':
		if max(handle.data['raw']['power']) <= ops.SPECTRAL_POWER_THRESHOLD: return length
		counter = length-1
		while True:
			if handle.data['raw']['power'][counter] > ops.SPECTRAL_POWER_THRESHOLD: break
			counter -= 1
		return counter+1



def delta(vals):
	output = np.diff(vals)
	output = np.insert(output, -1, 0)
	return output


def odf(data, numberAvg): 
	# with Norbert Schnell
	# 	an averaged delta function
	dataLen = len(data)
	matrix = np.zeros((numberAvg,dataLen)) # matrix of length x number average
	for frame in range(len(data)):
		for i in range(1,numberAvg+1):
			if frame-i >= 0: matrix[i-1][frame] = data[frame-i] # fill it
	medians = np.median(matrix, axis=0)
	return data-medians




#############
## HELPERS ##
#############
#def getSubBandSlices(binFreqs):
#	subBandsHz = [[50, 400], [400, 800], [800, 2000], [2000, 12000]]
#	outputSlices = []
#	for band in subBandsHz:
#		low = np.argmin(np.abs(binFreqs-band[0]))
#		high = np.argmin(np.abs(binFreqs-band[1]))
#		outputSlices.append(slice(low, high))
#	return outputSlices
#	
#
#
#
#def fluxAndSubBandedFlux(FFTs, FLUX_LAST_FFT_FRAME, subBandSlices):
#	# un-normalised flux
#	FFTLen = FFTs.shape[1]
#	FrameLen = FFTs.shape[0]
#	subBandSlices.insert(0, slice(0, FFTLen))
#	if FLUX_LAST_FFT_FRAME == None: FLUX_LAST_FFT_FRAME = np.zeros(FFTLen)
#	output = np.empty((len(subBandSlices), FrameLen))
#	for bidx, b in enumerate(subBandSlices):
#		# flux of last frame
#		output[bidx][0] = np.sum(np.power((FFTs[0][b]-FLUX_LAST_FFT_FRAME[b]), 2)) 
#		# flux between frames
#		for i in range(1, FrameLen):
#			output[bidx][i] = np.sum(np.power((FFTs[i][b]-FFTs[i-1][b]), 2)) 
#	return output
#
#
#def windowsStats(windows, handle, ag):
#	output = {}
#	output['mean'] = np.mean(windows, axis=1)
#	output['std'] = np.std(windows, axis=1)
#	output['sum'] = np.sum(windows, axis=1)
#	output['max'] = np.amax(windows, axis=1)
#	output['fluxes'] = fluxAndSubBandedFlux(windows, handle.FLUX_LAST_FFT, ag.subBandSlices[:])
#	return output
#
#
#def removeNans(output, val):
#	output[np.isnan(output)] = val # replace nans
#	output[np.isinf(output)] = val # replace infs
#	return output
#
#
#def getBand(twoDArray, whichBand): # returns only one vertical band of DxN matrix
#	return twoDArray.take([whichBand],axis=1).transpose()[0]
#
#
def hannWin(size):
	n = np.arange(size)
	w = 0.5*(1.0-np.cos(2*np.pi*n/(float(size-1))))
	return w



########################
## DESCRIPTOR MIXTURE ##
########################
def mixTwoTimeVaryingFeatures(descript, tgtTime, mixDur, mixPointer, srcPointer):
	if descript.find('power') != -1 or descript.find('rms') != -1: # powers are summed, not averaged
		# (e1+e2)
		#print "adding powers"
		#print "\t", mixPointer[descript][tgtTime:tgtTime+mixDur]
		#print "\t", srcPointer[descript][0:mixDur]
		#print "\t=", mixPointer[descript][tgtTime:tgtTime+mixDur]+srcPointer[descript][0:mixDur]
		return mixPointer[descript][tgtTime:tgtTime+mixDur]+srcPointer[descript][0:mixDur]
	else:
		# ((d1*e1)+(d1*e1)) / (e1+e2)
#		print "mixing descriptors"
#		print "time", tgtTime, "mix duration", mixDur, len(mixPointer[descript]), len(srcPointer[descript])
#		print "\t", mixPointer[descript][tgtTime:tgtTime+mixDur]
#		print "\t", srcPointer[descript][0:mixDur]
#		print "\t=", descript, ((mixPointer[descript][tgtTime:tgtTime+mixDur]*mixPointer['power'][tgtTime:tgtTime+mixDur])+(srcPointer[descript][0:mixDur]*srcPointer['power'][0:mixDur]))/(mixPointer['power'][tgtTime:tgtTime+mixDur]+srcPointer['power'][0:mixDur])
		return ((mixPointer[descript][tgtTime:tgtTime+mixDur]*mixPointer['power'][tgtTime:tgtTime+mixDur])+(srcPointer[descript][0:mixDur]*srcPointer['power'][0:mixDur]))/(mixPointer['power'][tgtTime:tgtTime+mixDur]+srcPointer['power'][0:mixDur])


def mixTwoSegmentedFeatures(descript, feature1, power1, feature2, power2):
	#print descript, feature1, power1, feature2, power2
	if descript in ['power-seg', 'rms-seg']:
		return power1+power2
	else: # weighted average
		return ((feature1*power1)+(feature2*power2))/(power1+power2)




#############################
## PACKAGES OF DESCRIPTORS ##
#############################
def isPackageDescriptor(name, listPackages):
	isPackage = False
	for desc in listPackages:
		if name.find(desc) != -1:
			isPackage = True
			break
	return isPackage




def name2packageExpansion(name, weight, neededBy):
	metricsToWrite = []
	packagename = name
	if name.find('mfccEnv') != -1 or name.find('mfccFrq') != -1:
		mfccWeightMask = [1, 0.95, 0.9, 0.78, 0.7, 0.65, 0.5, 0.3, 0.2]
		nBands = 8
		if name.find('mfccEnv') != -1: startBand = 1
		elif name.find('mfccFrq') != -1: startBand = int(ops.DESCRIPTOR_NUMB_BANDS/2.0)+1
		for i in range(nBands): metricsToWrite.append(('mfcc'+str(i+startBand)+name[7:], weight*(mfccWeightMask[i]/sum(mfccWeightMask)), neededBy ))
	elif name.find('chromas') != -1:
		for i in range(12): metricsToWrite.append( ( 'chroma'+str(i)+name[9:], weight/12., neededBy ) )
	elif name.find('cluster_cents') != -1:
		numb = len(ops.DESCRIPTOR_CLUSTS_DESCRIPTOR_DIM)
		for i in range(numb):
			metricsToWrite.append( ( 'cluster_cent'+str(i)+name[13:], weight/float(numb), neededBy ) )
		#print metricsToWrite
	elif name.find('peakfrqs') != -1:
		for i in range(ops.DESCRIPTOR_NUMB_PEAKS):
			metricsToWrite.append( ('peakfrq'+str(i)+name[10:], weight*((ops.DESCRIPTOR_NUMB_PEAKS-i)/float(ops.DESCRIPTOR_NUMB_PEAKS)), neededBy ) )
			metricsToWrite.append( ('peakamp'+str(i)+name[10:], 0, neededBy) )
#	elif name.find('loudestpeak') != -1:
#		metricsToWrite.append( ('peakfrq0'+name[10:], weight*((ops.DESCRIPTOR_NUMB_PEAKS-i)/float(ops.DESCRIPTOR_NUMB_PEAKS)), neededBy ) )
#		metricsToWrite.append( ('peakamp0'+name[10:], 0, neededBy) )
#		for i in range(ops.DESCRIPTOR_NUMB_PEAKS+1):
#			metricsToWrite.append( ('peakfrq'+str(i)+name[10:], weight*((ops.DESCRIPTOR_NUMB_PEAKS-i)/float(ops.DESCRIPTOR_NUMB_PEAKS)), ['target'] ) )
#			metricsToWrite.append( ('peakamp'+str(i)+name[10:], 0, ['target']) )
	elif name.find('crests') != -1:
		for i in range(4):
			metricsToWrite.append( [ 'crest'+str(i)+name[6:], weight/4., neededBy ] )
	elif name.find('flatnesses') != -1:
		for i in range(4):
			metricsToWrite.append( [ 'flatness'+str(i)+name[10:], weight/4., neededBy ] )
	elif name.find('harmonictristimuluses') != -1:
		for i in range(3):
			metricsToWrite.append( [ 'harmonictristimulus'+str(i)+name[21:], weight/3., neededBy ] )
	elif name.find('perceptualtristimuluses') != -1:
		for i in range(3):
			metricsToWrite.append( [ 'perceptualtristimulus'+str(i)+name[23:], weight/3., neededBy ] )
	else:
		metricsToWrite = [(name, weight, neededBy)]
		packagename = None
	return packagename, metricsToWrite




##############################
## AUDIOGUIDE FUNCTION CALL ##
##############################
def analSegment(d, handle, start, end, typed, ag):
	global ALL_SEG_METHDS
	# if already computed as init, skip it! -- really a hack in case there is a bug -- shouldn't happen
	if typed != "Recalc" and start == None and end == None and handle.data['raw'].has_key(d.name): pass
	#print d.name, d.seg
	if d.seg:
		# SINGLE-NUMBER METHODS
		if d.name == "dur-seg": # raw segment duration
			output = [handle.lengthInFrames]
		elif d.name == "effDur-seg": # perceived segment duration
			output = [effectiveDur(handle, start, ag)]
		elif d.name == "MIDIPitch-seg":
			output = [MIDIPitchByFileName(handle.name, handle.midiPitchMethod, handle)]
		elif d.name == "peakTime-seg":
			output = peakTimeSeg(handle.data['raw']['power'][start:end])
		elif d.name == "percentageInFile-seg":
			output = percentageInFile(handle, start, end)
		# DESCRIPTOR SLOPE
		elif d.type == 'slope-regression':
			output = descriptorSlope(handle.data['raw'], d.parents[0], start)
		# FLAT MEAN - NOT ENERGY WEIGHTED
		elif d.type  == 'mean':
			output = [np.average(handle.data['raw'][d.parents[0]][start:end])]
		# SEGMENT-AVERAGE METHODS
		elif d.seg_method  == 'max':
			output = [np.max(handle.data['raw'][d.parents[0]][start:end])]
		elif d.seg_method  == 'median':
			output = [np.median(handle.data['raw'][d.parents[0]][start:end])]
		else: # 'weighted_mean'
			try:
				output = [np.average(handle.data['raw'][d.parents[0]][start:end], weights=handle.data['raw']['power'][start:end])]
			except ZeroDivisionError: # in case all powers are zero
				output = [np.average(handle.data['raw'][d.parents[0]][start:end], weights=np.clip(handle.data['raw']['power'][start:end], 0.0001, sys.maxint))]

	else:
		# DELTA
		if d.type == 'delta':
			output = delta(handle.data['raw'][d.parents[0]][start:end])
			print output
		# DELTA DELTA
		elif d.type == 'deltadelta':
			output = delta(handle.data['raw'][d.parents[0]][start:end])
			output = delta(output)
		# ONSET DETECTION FUNCTION CALL
		elif d.type == 'onsetdetection':
			analSplit = d.name.split('-')
			output = odf(handle.data['raw'][d.parents[0]][start:end], int(analSplit[2]))
		elif d.name == "loudestpeak":
			output = handle.data['raw']['peakfrq0'][start:end]
		else:
			print "No method for", anal, ", Quitting..."
			utilities.exit()
	return output


def multiSegAnal(handle, descript, start, end):
	if start == None: start = 0
	p = int(handle.data['raw']['peakTime-seg'][start])
	output = np.zeros(3)
	output[0] = [simcalc.energyWeightSingle((handle.data['raw'][descript][start:p-1]), handle.data['raw']['power'][start:p-1])][0]
	output[1] = [simcalc.energyWeightSingle((handle.data['raw'][descript][p:p+1]), handle.data['raw']['power'][p:p+1])][0]
	output[2] = [simcalc.energyWeightSingle((handle.data['raw'][descript][p+2:end]), handle.data['raw']['power'][p+2:end])][0]
	return output
